# -*- Mode: python; coding: utf-8; tab-width: 4; indent-tabs-mode: nil; -*-
#
# GNOME Activity Journal
#
# Copyright © 2009-2010 Seif Lotfy <seif@lotfy.com>
# Copyright © 2010 Siegfried Gevatter <siegfried@gevatter.com>
# Copyright © 2010 Randal Barlow <email.tehk@gmail.com>
# Copyright © 2011 Stefano Candori <stefano.candori@gmail.com>
# Copyright © 2020 The GNOME Activity Journal developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import gettext
import time
import datetime
import os
import sys

import gi
from gi.repository import GObject, GLib, Gtk, Gdk, GdkPixbuf, Gio

from .activity_widgets import MultiViewContainer, TimelineViewContainer, ThumbViewContainer
from .supporting_widgets import DayButton, DayLabel, Toolbar, SearchBox, ContextMenu, ThrobberPopupButton, \
ContextMenuMolteplicity, NiceBar, AboutDialog, VideoPreviewTooltip, AudioPreviewTooltip
from .histogram import HistogramWidget
from .store import Store, tdelta, STORE, CLIENT
from .config import settings, preference_settings, get_icon_path, get_data_path, PluginManager
from .Indicator import TrayIconManager
from .common import *
from .blacklist import BLACKLIST
from . import config

#TODO granularity scrool, alignment timelineview_icon
#more metadata? use website cache as thumbpreview??
class ViewContainer(Gtk.Notebook):
    __gsignals__ = {
        "new-view-added" : (GObject.SignalFlags.RUN_LAST, None,(GObject.TYPE_PYOBJECT,)),
        "view-button-clicked" : (GObject.SignalFlags.RUN_LAST, None,(GObject.TYPE_PYOBJECT,GObject.TYPE_INT)),
    }

    class ViewStruct(object):
        view = None
        button = None

        def __init__(self, view, button):
            self.view = view
            self.button = button

    def __init__(self, store):
        super(ViewContainer, self).__init__()
        self.store = store
        self.set_show_tabs(False)
        self.set_show_border(False)
        self.pages = []
        self.tool_buttons = []
#        self.button_group = []

    def set_day(self, day, page=None):
        force_update = True
        if page == None:
            page = self.get_current_page()
            force_update = False
        if hasattr(self.pages[page], "set_day"):
            self.pages[page].set_day(day, self.store, force_update)

    def set_zoom(self, zoom):
        page = self.get_current_page()
        if hasattr(self.pages[page], "set_zoom"):
            self.pages[page].set_zoom(zoom)

    def _register_new_view(self, viewstruct):
        self.append_page(viewstruct.view)
        self.pages.append(viewstruct.view)
        self.tool_buttons.append(viewstruct.button)
        n = len(self.tool_buttons)
#        if n == 1:
#            self.button_group = viewstruct.button.get_group()
#        elif n > 1:
#            viewstruct.button.set_group(self.button_group)
        viewstruct.button.set_can_focus(True)
        viewstruct.button.connect("toggled", self.view_button_toggled, len(self.pages)-1)
        viewstruct.view.show_all()
        return self.pages.index(viewstruct.view)

    def register_new_view(self, viewstruct):
        i = self._register_new_view(viewstruct)
        self.emit("new-view-added", viewstruct)
        return i

    def remove_view(self, i, destroy=False):
        """
        :param i: a page number index starting at zero
        """
        tb = self.tool_buttons[i]
        page = self.pages[i]
        del self.pages[i]
        del self.tool_buttons[i]
        self.remove_page(i)
        tb.parent.remove(tb)
        if destroy:
            page.destroy()
            tb.destroy()

    @property
    def page(self):
        return self.get_current_page()

    @property
    def name(self):
        return self.pages[self.page].name

    def view_button_toggled(self, button, i):
        if not button.get_active():return
        button.grab_focus()
        self.emit("view-button-clicked", button, i)

    def set_view_page(self, i):
        self.set_current_page(i)

    def _register_default_view(self, view):
        toolbutton = Toolbar.get_toolbutton(view.icon_path, view.dsc_text)
        self._register_new_view(self.ViewStruct(view, toolbutton))
        self.set_view_page(0)

    def set_zoom_slider(self, hscale):
        #FIXME dirty hack
        for page in self.pages:
            page.set_slider(hscale)

    def toggle_erase_mode(self):
        for page in self.pages:
            page.toggle_erase_mode()


class PortalWindow(Gtk.ApplicationWindow):
    """
    The primary application window
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.application = kwargs["application"]

        p = Gtk.CssProvider.new()
        s = Gdk.Screen.get_default()
        p.load_from_path(get_data_path("style.css"))
        Gtk.StyleContext.add_provider_for_screen(s, p, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION)

        # write to gsettings from current theme
        update_default_search_colors()

        # Important
        self.date = None
        self.paused = False
        self._request_size()
        self.store = STORE
        self.day_iter = self.store.today
        self.pages_loaded = 0
        self.current_zoom = settings["zoom-level"]
        self.thumb_view_zoom = settings["thumb-view-zoom-level"]
        self.timeline_view_zoom = settings["timeline-view-zoom-level"]
        self.in_erase_mode = False
        self.view = ViewContainer(self.store)
        self.toolbar = Toolbar()
        default_views = (MultiViewContainer(), ThumbViewContainer(), TimelineViewContainer())
        default_views[0].connect("view-ready", self._on_view_ready)
        # FixMe: change all 'nolazy' maps to normal loops
        nolazy = list(map(self.view._register_default_view, default_views))
        nolazy = list(map(self.toolbar.add_new_view_button, self.view.tool_buttons))
        ContextMenu.set_parent_window(self)
        ContextMenuMolteplicity.set_parent_window(self)
        self.histogram = HistogramWidget()
        self.histogram.set_store(self.store)
        self.backward_button, ev_backward_button = DayButton.new(0)
        self.forward_button, ev_forward_button = DayButton.new(1, sensitive=False)
        self.nicebar = NiceBar()
        self.nicebar_timeout = None
        self.spinner = None
        self.use_spinner = True

        # use a table for the spinner (otherwise the spinner is massive!)
        spinner_table = Gtk.Table(3, 3, False)
        label = Gtk.Label()
        label.set_markup(_("<span size=\"x-large\"><b>Loading Journal...</b></span>"))
        vbox = Gtk.VBox(False, 5)

        if self.use_spinner:
            self.spinner = Gtk.Spinner()
            self.spinner.start()
            vbox.pack_start(self.spinner, False, False, 0)
        else:
            pix = GdkPixbuf.Pixbuf.new_from_file(get_data_path("zeitgeist-logo.svg"))
            pix = pix.scale_simple(200, 200, GdkPixbuf.InterpType.BILINEAR)
            zlogo = Gtk.Image.new_from_pixbuf(pix)
            vbox.pack_start(zlogo, False, False, 0)

        vbox.pack_start(label, True, True, 0)
        spinner_table.attach(vbox, 1, 2, 1, 2, Gtk.AttachOptions.EXPAND, Gtk.AttachOptions.EXPAND)

        self.hscale_range_start = 0.0
        self.hscale_range_end = 3.0
        self.hscale = Gtk.Scale.new_with_range(Gtk.Orientation.HORIZONTAL,
                                               self.hscale_range_start,
                                               self.hscale_range_end,
                                               1.0)
        self.load_zoom()
        self.hscale.set_size_request(120, -1)
        self.hscale.set_draw_value(False)

        al = Gtk.Alignment.new(0.5, 0.5, 1.0, 1.0)
        al.add(self.hscale)

        self.btn_zoom_in = btn_zoom_in = Gtk.Button.new_from_icon_name("zoom-in-symbolic", Gtk.IconSize.MENU)
        btn_zoom_in.connect("clicked", self._on_zoom_in_button_clicked)
        btn_zoom_in.set_relief(Gtk.ReliefStyle.NONE)

        self.btn_zoom_out = btn_zoom_out =  Gtk.Button.new_from_icon_name("zoom-out-symbolic", Gtk.IconSize.MENU)
        btn_zoom_out.connect("clicked", self._on_zoom_out_button_clicked)
        btn_zoom_out.set_relief(Gtk.ReliefStyle.NONE)

        self.throbber_popup_button = ThrobberPopupButton()

        # Search button
        self.ignore_search_toggle = False
        self.search_button = sb = Gtk.ToggleButton()
        image = Gtk.Image.new_from_icon_name("edit-find-symbolic", Gtk.IconSize.BUTTON)
        self.search_button.set_image(image)
        self.search_button.connect("toggled", self.toggle_searchbox_visibility)

        # search bar
        self.search_dialog = SearchBox
        self.search_bar = self.create_search_bar()
        self.search_bar_revealed = False
        self.search_bar.connect("notify::search-mode-enabled", self.search_bar_mode_enabled_changed)
        self.search_bar.show_all()
        self.search_dialog.connect("matching-day", self.set_matching_day)

        # Widget placement
        vbox = Gtk.VBox(); hbox = Gtk.HBox();self.scale_box = Gtk.HBox();self.histogramhbox = Gtk.HBox()
        vbox_general = Gtk.VBox();
        self.toolbar_box = Gtk.HBox()
        self.scale_menu_box = Gtk.HBox()
        hbox.pack_start(ev_backward_button, False, False, 0); hbox.pack_start(self.view, True, True, 6)
        hbox.pack_end(ev_forward_button, False, False, 0);

        self.scale_box.set_valign(Gtk.Align.CENTER)
        self.scale_box.pack_start(btn_zoom_out, False, False, 0);
        self.scale_box.pack_start(al, False, False, 0)
        self.scale_box.pack_start(btn_zoom_in, False, False, 0);

        self.toolbar_box.pack_start(self.toolbar, False, True, 0);
        self.scale_menu_box.pack_end(self.throbber_popup_button, False, False, 6);
        self.scale_menu_box.pack_end(self.search_button, False, False, 6);
        self.scale_menu_box.pack_end(self.scale_box, False, False, 6);

        self.histogramhbox.pack_end(self.histogram, True, True, 32);
        self.histogramhbox.set_sensitive(False)

        vbox.pack_start(self.search_bar, False, False, 0);
        vbox.pack_start(self.nicebar, False, False, 0);
        vbox.pack_start(hbox, True, True, 5);
        vbox.pack_end(self.histogramhbox, False, False, 0)

        self.spinner_notebook = Gtk.Notebook()
        self.spinner_notebook.set_show_tabs(False)
        self.spinner_notebook.set_show_border(False)
        self.spinner_notebook.append_page(spinner_table)
        self.spinner_notebook.append_page(vbox)
        vbox_general.pack_start(self.spinner_notebook, False, True, 0)
        self.add(vbox_general)
        vbox_general.show_all()
        self.scale_box.hide()

        self.header_bar = Gtk.HeaderBar.new()
        self.header_bar.pack_start(self.toolbar_box)
        self.header_bar.pack_end(self.scale_menu_box)
        self.header_bar.set_show_close_button(True)
        self.set_titlebar(self.header_bar)
        self.header_bar.show_all()
        self.scale_box.hide()

        self.show()
        self.nicebar.hide()
        SearchBox.hide()

        #Tray Icon
        self.tray_manager = TrayIconManager(self)
        # Settings
        self.view.set_day(self.store.today)
        self.view.set_zoom_slider(self.hscale)
        # Signal connections
        self.view.connect("new-view-added", lambda w, v: self.toolbar.add_new_view_button(v.button, len(self.view.tool_buttons)))
        self.connect("destroy", self.quit)
        self.connect("delete-event", self.on_delete)
        self.toolbar.connect("previous", self.previous)
        self.toolbar.connect("jump-to-today", lambda w: self.set_date(datetime.date.today()))
        self.toolbar.connect("jump-to-selected-day", self.jump_to_selected_day)
        self.toolbar.connect("next", self.next)
        self.hscale.connect("value-changed", self._on_zoom_changed)
        self.backward_button.connect("clicked", self.previous)
        self.forward_button.connect("clicked", self.next)
        self.forward_button.connect("jump-to-today", lambda w: self.set_date(datetime.date.today()))
        self.histogram.connect("date-changed", lambda w, date: self.set_date(date))
        self.view.connect("view-button-clicked", self.on_view_button_click)
        self.store.connect("update", self.histogram.histogram.set_store)
        SearchBox.connect("search", self._on_search)
        SearchBox.connect("clear", self._on_search_clear)

        # Window configuration
        self.set_icon_name("org.gnome.ActivityJournal")
        self.set_icon_list(
            [GdkPixbuf.Pixbuf.new_from_file(get_icon_path(f)) for f in (
                "hicolor/16x16/apps/org.gnome.ActivityJournal.png",
                "hicolor/24x24/apps/org.gnome.ActivityJournal.png",
                "hicolor/32x32/apps/org.gnome.ActivityJournal.png",
                "hicolor/48x48/apps/org.gnome.ActivityJournal.png",
                "hicolor/256x256/apps/org.gnome.ActivityJournal.png")])
        GObject.idle_add(self.setup)
        GObject.idle_add(self.load_plugins)

        # set window title as per journal logging state
        enabled = BLACKLIST.get_incognito()
        self.update_title_from_status(enabled)

        self.init_actions()

    def init_actions(self):
        action = Gio.SimpleAction.new("toggle_erase_mode", None)
        action.connect("activate", self.on_toggle_erase_mode)
        self.add_action(action)

    def search_bar_mode_enabled_changed(self, a, b):
        if not a.props.search_mode_enabled:
            self.search_button.set_active(False)

    def on_key_pressed(self, window, event, search_bar):
        search_bar_revealed = search_bar.handle_event(event)
        if search_bar_revealed:
            self.toggle_search_bar(True)
        return search_bar_revealed

    def set_searchbutton_state(self, state):
        self.ignore_search_toggle = True
        search_button.set_active(True)
        self.ignore_search_toggle = False

    def toggle_searchbox_visibility(self, w):
        if self.ignore_search_toggle:
            return

        show = self.search_button.get_active()
        self.toggle_search_bar(show)

    def toggle_search_bar(self, show=True):
        self.search_bar.set_search_mode(show)
        self.search_button.set_active(show)

        if show:
            self.search_bar.show_all()

    def create_search_bar(self):
        search_bar = Gtk.SearchBar.new()
        hbox = Gtk.HBox.new(False, 6)

        search_bar.add(hbox)
        hbox.pack_start(self.search_dialog, True, True, 0)
        search_bar.connect_entry(self.search_dialog.search)

        self.connect("key-press-event", self.on_key_pressed, search_bar)

        return search_bar

    def load_plugins(self):
        # rewrite plugin using peas gtk
        return False

    def setup(self, *args):
        self.set_title_from_date(self.day_iter.date)
        self.histogram.set_dates(self.active_dates)
        self.histogram.scroll_to_end()
        return False

    def set_visibility(self, val):
        if val: self.show()
        else: self.hide()

    def toggle_visibility(self):
        if self.get_property("visible"):
            self.hide()
            return False
        self.show()
        return True

    @property
    def active_dates(self):
        date = self.day_iter.date
        if self.view.page != 0: return [date]
        dates = []
        for i in range(self.view.pages[0].num_pages):
            dates.append(date + tdelta(-i))
        dates.sort()
        return dates

    def set_day(self, day):
        self.toolbar.sync_date(day.date)
        self.day_iter = day
        self.handle_button_sensitivity(day.date)
        self.view.set_day(day)
        self.histogram.set_dates(self.active_dates)
        self.set_title_from_date(day.date)

    def set_date(self, date):
        self.set_day(self.store[date])

    def set_matching_day(self, searchbox, date):
        self.set_date(date)

    def next(self, *args):
        day = self.day_iter.next(self.store)
        self.set_day(day)

    def set_today(self):
        self.set_date(datetime.date.today())
        return False

    def jump_to_selected_day(self, toolbar, date):
        today = datetime.date.today()
        date = datetime.date(date.year, date.month + 1, date.day)

        if date > today:
            GLib.timeout_add(100, self.set_today)
            return

        self.set_day(self.store[date])

    def previous(self, *args):
        day = self.day_iter.previous(self.store)
        self.set_day(day)

    def handle_button_sensitivity(self, date):
        today = datetime.date.today()
        if date == today:
            self.forward_button.set_sensitive(False)
            self.toolbar.nextd_button.set_sensitive(False)
            self.toolbar.home_button.set_sensitive(False)
        else:
            self.forward_button.set_leading(True)
            self.forward_button.set_sensitive(True)
            self.toolbar.nextd_button.set_sensitive(True)
            self.toolbar.home_button.set_sensitive(True)

    def on_view_button_click(self, w, button, i):
        self.view.set_view_page(i)
        self.view.set_day(self.day_iter, page=i)
        self.histogram.set_dates(self.active_dates)
        self.set_title_from_date(self.day_iter.date)
        if i == 0:
            self.scale_box.hide()
        else:
            zoom_value = self.load_zoom()
            self.view.set_zoom(zoom_value)
            self.scale_box.show()

    def load_zoom(self):
        view = self.view.name

        if not config.SEPARATE_ZOOM_PER_VIEW:
            zoom_value = self.current_zoom
        else:
            if view == "thumb-view":
                zoom_value = self.thumb_view_zoom
            elif view == "timeline-view":
                zoom_value = self.timeline_view_zoom
            elif view == "multi-view":
                zoom_value = -1
            else:
                raise Exception("Unsupported view: %s" % view)

        if zoom_value >= 0:
            self.hscale.set_value(zoom_value)

        return zoom_value

    def _switch_page(self, view):
        if self.pages_loaded == view.num_pages - 1:
            self.histogramhbox.set_sensitive(True)
            self.spinner.stop()
            self.spinner_notebook.set_current_page(1)
        else:
            self.pages_loaded += 1

        return False

    def _on_view_ready(self, view):
        GLib.idle_add (self._switch_page, view)

    def _on_search(self, box, results):
        dates = []
        for obj in results:
            dates.append(datetime.date.fromtimestamp(int(obj.event.timestamp)/1000.0))
        self.histogram.histogram.set_highlighted(dates)

    def _on_search_clear(self, *args):
        self.histogram.histogram.clear_highlighted()
        self.set_today()

    def _request_size(self):
        screen = self.get_screen().get_monitor_geometry(
            self.get_screen().get_monitor_at_point(*self.get_position()))
        min_size = (1024, 600) # minimum netbook size
        size = [
            min(max(int(screen.width * 0.80), min_size[0]), screen.width),
            min(max(int(screen.height * 0.75), min_size[1]), screen.height)
        ]

        if settings["window-width"] and settings["window-width"] <= screen.width:
            size[0] = settings['window-width']
        if settings["window-height"] and settings["window-height"] <= screen.height:
            size[1] = settings["window-height"]

        geometry = Gdk.Geometry()
        geometry.min_width = 1024
        geometry.min_height = 360

        self.set_geometry_hints(None, geometry, Gdk.WindowHints.MIN_SIZE)
        self.resize(size[0], size[1])
        self._requested_size = size

    def save_zoom(self, zoom_value):
        view = self.view.name

        if not config.SEPARATE_ZOOM_PER_VIEW:
            self.current_zoom = zoom_value
            zoom_key = "zoom-level"
        else:
            if view == "thumb-view":
                self.thumb_view_zoom = zoom_value
                zoom_key = "thumb-view-zoom-level"
            elif view == "timeline-view":
                self.timeline_view_zoom = zoom_value
                zoom_key = "timeline-view-zoom-level"
            else:
                raise Exception("Unsupported view: %s" % view)

        settings[zoom_key] = zoom_value

    def _on_zoom_changed(self, hscale):
        zoom_value = int(hscale.get_value())
        self.view.set_zoom(zoom_value)

        self.save_zoom(zoom_value)

        zoom_out = (self.hscale.get_value() > self.hscale_range_start)
        self.btn_zoom_out.set_sensitive(zoom_out)
        zoom_in = (self.hscale.get_value() < self.hscale_range_end)
        self.btn_zoom_in.set_sensitive(zoom_in)

    def _adjust_scale_value(self, delta):
        current_zoom = int(self.hscale.get_value())
        self.hscale.set_value(current_zoom + delta)

    def _on_zoom_out_button_clicked(self, button):
        if self.hscale.get_value() > self.hscale_range_start:
            self._adjust_scale_value(-1)

    def _on_zoom_in_button_clicked(self, button):
        if self.hscale.get_value() < self.hscale_range_end:
            self._adjust_scale_value(1)

    def on_toggle_erase_mode(self, *args):
        self.in_erase_mode = not self.in_erase_mode
        if self.in_erase_mode:
            if self.nicebar_timeout:
                GObject.source_remove(self.nicebar_timeout)
                self.nicebar_timeout = None
            message = _("Erase Mode is active")
            background = self.nicebar.ALERTBACKGROUND
        else:
            message = _("Erase Mode deactivated")
            background = self.nicebar.NORMALBACKGROUND
            self.nicebar_timeout = GObject.timeout_add(3000, self.nicebar.remove_message, self)

        self.nicebar.display_message(message, background=background)
        self.view.toggle_erase_mode()

    def set_title_from_date(self, date):
        self.date = date
        pages = self.view.pages[0].num_pages
        if self.view.page == 0:
            start_date = date + tdelta(-pages+1)
        else:
            start_date = date
        if date == datetime.date.today():
            end = _("Today")
            start = start_date.strftime("%A")
        elif date == datetime.date.today() + tdelta(-1):
            end = _("Yesterday")
            start = start_date.strftime("%A")
        elif date + tdelta(6) > datetime.date.today():
            end = date.strftime("%A")
            start = start_date.strftime("%A")
        else:
            start = start_date.strftime("%d %B")
            end = date.strftime("%d %B")
        if self.view.page != 0:
            title = end + " - " + _("Activity Journal")
        else:
            title = _("%(start)s to %(end)s") % ({ 'start': start, 'end': end}) + " - " + _("Activity Journal")

        if self.paused:
            title += " " +  _("(Paused)")
        self.set_title (title)

    def update_title_from_status(self, paused):
        self.paused = paused
        if not self.date:
            return

        self.set_title_from_date(self.date)

    def on_delete(self, w, event):
        x, y = self.get_size()
        settings["window-width"] = x
        settings["window-height"] = y
        if preference_settings["enable-tray-icon"]:
            self.set_visibility(False)
            return True

    def quit_and_save(self, *args):
        x, y = self.get_size()
        settings["window-width"] = x
        settings["window-height"] = y
        self.application.quit()

    def quit(self, *args):
        self.application.quit()

class Application(Gtk.Application):
    def __init__(self, *args, **kwargs):
        super().__init__(
            *args,
            application_id="org.gnome.ActivityJournal",
            flags=Gio.ApplicationFlags.HANDLES_COMMAND_LINE,
            **kwargs
        )
        self.window = None

        self.add_main_option(
            "debug",
            ord("d"),
            GLib.OptionFlags.NONE,
            GLib.OptionArg.NONE,
            "Print debug information",
            None,
        )

        self.add_main_option(
            "version",
            ord("v"),
            GLib.OptionFlags.NONE,
            GLib.OptionArg.NONE,
            "Print version information",
            None,
        )

    def do_startup(self):
        Gtk.Application.do_startup(self)

        action = Gio.SimpleAction.new("preferences", None)
        action.connect("activate", self.on_preferences)
        self.add_action(action)

        action = Gio.SimpleAction.new("about", None)
        action.connect("activate", self.on_about)
        self.add_action(action)

        action = Gio.SimpleAction.new("quit", None)
        action.connect("activate", self.on_quit)
        self.add_action(action)

        self.set_accels_for_action("app.quit", ["<Primary>q"])

    def do_activate(self):
        # We only allow a single window and raise any existing ones
        if not self.window:
            # Windows are associated with the application
            # when the last one is closed the application shuts down
            self.window = PortalWindow(application=self, title="Main Window")

        self.window.present()

    def do_command_line(self, command_line):
        options = command_line.get_options_dict()
        # convert GVariantDict -> GVariant -> dict
        options = options.end().unpack()

        if "debug" in options:
            config.DEBUG = options["debug"]

        if "version" in options:
            print("%s %s" % (config.NAME, config.VERSION))
            sys.exit(0)

        self.activate()
        return 0

    def on_about(self, action, param):
        about_dialog = AboutDialog(transient_for=self.window, modal=True)
        about_dialog.run()
        about_dialog.destroy()

    def on_preferences(self, action, param):
        pref = Preferences(self.window)
        pref.show()

    def on_quit(self, action, param):
        self.window.quit_and_save()
        self.quit()

class Preferences(object):
    def __init__(self, window):
        self.window = window
        self.load_widgets()

    def load_widgets(self):
        builder = Gtk.Builder.new_from_file(get_data_path("ui/preferences.ui"))

        # dialog
        self.preferences_dialog = builder.get_object("preferences_dialog")

        # info bar
        self.info_bar = info_bar = builder.get_object("info_bar")
        info_bar.connect("response", self.on_info_bar_response)
        self.info_bar_label = label = Gtk.Label()
        content_area = info_bar.get_content_area()
        content_area.add(label)
        self.info_bar.hide()

        # separate zoom per view check button
        self.separate_zoom_per_view_check_button = builder.get_object("use_separate_zoom_levels_checkbutton")
        self.separate_zoom_per_view_check_button.connect("toggled", self.separate_zoom_per_view_check_button_toggled)
        preference_settings.bind("separate-zoom-per-view", self.separate_zoom_per_view_check_button, "active", Gio.SettingsBindFlags.DEFAULT)

        # enable tray icon check button
        self.enable_tray_icon_check_button = builder.get_object("enable_tray_icon_checkbutton")
        self.enable_tray_icon_check_button.connect("toggled", self.enable_tray_icon_check_button_toggled)
        preference_settings.bind("enable-tray-icon", self.enable_tray_icon_check_button, "active", Gio.SettingsBindFlags.GET)

        # delete confirm in erase mode check button
        self.delete_confirm_check_button = builder.get_object("delete_confirm_checkbutton")
        self.delete_confirm_check_button.connect("toggled", self.delete_confirm_check_button_toggled)
        preference_settings.bind("confirm-item-deletion", self.delete_confirm_check_button, "active", Gio.SettingsBindFlags.DEFAULT)

        # hide empty day labels check button
        self.hide_empty_day_labels_check_button = builder.get_object("hide_empty_day_labels_checkbutton")
        preference_settings.bind("hide-empty-day-labels", self.hide_empty_day_labels_check_button, "active", Gio.SettingsBindFlags.DEFAULT)
        self.hide_empty_day_labels_check_button.connect("toggled", self.hide_empty_day_labels_check_button_toggled)

        # search in subject uri also
        self.search_in_subject_uri_check_button = builder.get_object("search_in_subject_uri_checkbutton")
        preference_settings.bind("search-in-subject-uri", self.search_in_subject_uri_check_button, "active", Gio.SettingsBindFlags.DEFAULT)
        self.search_in_subject_uri_check_button.connect("toggled", self.search_in_subject_uri_check_button_toggled)

        # search ignore single char
        self.search_ignore_single_char_check_button = builder.get_object("search_ignore_single_char_checkbutton")
        preference_settings.bind("search-ignore-single-char", self.search_ignore_single_char_check_button, "active", Gio.SettingsBindFlags.DEFAULT)
        self.search_ignore_single_char_check_button.connect("toggled", self.search_ignore_single_char_check_button_toggled)

        # search is case sensitive
        self.search_is_case_sensitive_check_button = builder.get_object("search_is_case_sensitive_checkbutton")
        preference_settings.bind("search-is-case-sensitive", self.search_is_case_sensitive_check_button, "active", Gio.SettingsBindFlags.DEFAULT)
        self.search_is_case_sensitive_check_button.connect("toggled", self.search_is_case_sensitive_check_button_toggled)

        # search results bold text
        self.search_results_use_bold_check_button = builder.get_object("search_results_use_bold_checkbutton")
        preference_settings.bind("search-results-use-bold-style", self.search_results_use_bold_check_button, "active", Gio.SettingsBindFlags.GET)
        self.search_results_use_bold_check_button.connect("toggled", self.search_results_use_bold_check_button_toggled)

        # search results large text
        self.search_results_use_large_check_button = builder.get_object("search_results_use_large_checkbutton")
        preference_settings.bind("search-results-use-large-style", self.search_results_use_large_check_button, "active", Gio.SettingsBindFlags.GET)
        self.search_results_use_large_check_button.connect("toggled", self.search_results_use_large_check_button_toggled)

        # default search color check button
        self.search_results_use_default_color_check_button = builder.get_object("search_results_use_default_color_checkbutton")
        preference_settings.bind("search-results-use-default-color", self.search_results_use_default_color_check_button, "active", Gio.SettingsBindFlags.GET)
        self.search_results_use_default_color_check_button.connect("toggled", self.search_results_use_default_color_check_button_toggled)

        use_default_search_colors = self.search_results_use_default_color_check_button.get_active()

        # custom text search results color
        text_rgba = get_gdk_rgba_from_string(config.SEARCH_RESULTS_TEXT_COLOR)
        self.search_results_text_color_button = builder.get_object("search_results_text_color_button")
        self.search_results_text_color_button.set_sensitive(not use_default_search_colors)
        self.search_results_text_color_button.set_property("rgba", text_rgba)
        self.search_results_text_color_button.connect("color-set", self.search_results_text_color_changed)

        # custom histogram search results color
        histogram_rgba = get_gdk_rgba_from_string(config.SEARCH_RESULTS_HISTOGRAM_COLOR)
        self.search_results_histogram_color_button = builder.get_object("search_results_histogram_color_button")
        self.search_results_histogram_color_button.set_property("rgba", histogram_rgba)
        self.search_results_histogram_color_button.set_sensitive(not use_default_search_colors)
        self.search_results_histogram_color_button.connect("color-set", self.search_results_histogram_color_changed)

        # audio preview check button
        self.show_audio_preview_check_button = builder.get_object("show_audio_preview_checkbutton")
        preference_settings.bind("show-audio-preview", self.show_audio_preview_check_button, "active", Gio.SettingsBindFlags.GET)
        self.show_audio_preview_check_button.connect("toggled", self.show_audio_preview_check_button_toggled)

        # video preview check button
        self.show_video_preview_check_button = builder.get_object("show_video_preview_checkbutton")
        preference_settings.bind("show-video-preview", self.show_video_preview_check_button, "active", Gio.SettingsBindFlags.GET)
        self.show_video_preview_check_button.connect("toggled", self.show_video_preview_check_button_toggled)

        # day part spin buttons
        self.early_morning_to_spin_button = builder.get_object("early_morning_to_spin_button")
        self.morning_to_spin_button = builder.get_object("morning_to_spin_button")
        self.afternoon_to_spin_button = builder.get_object("afternoon_to_spin_button")
        self.evening_to_spin_button = builder.get_object("evening_to_spin_button")

        preference_settings.bind("day-part-early-morning-to", self.early_morning_to_spin_button, "value", Gio.SettingsBindFlags.DEFAULT)
        preference_settings.bind("day-part-morning-to", self.morning_to_spin_button, "value", Gio.SettingsBindFlags.DEFAULT)
        preference_settings.bind("day-part-afternoon-to", self.afternoon_to_spin_button, "value", Gio.SettingsBindFlags.DEFAULT)
        preference_settings.bind("day-part-evening-to", self.evening_to_spin_button, "value", Gio.SettingsBindFlags.DEFAULT)

        self.early_morning_to_spin_button.connect("value-changed", self.day_part_spin_button_value_changed)
        self.morning_to_spin_button.connect("value-changed", self.day_part_spin_button_value_changed)
        self.afternoon_to_spin_button.connect("value-changed", self.day_part_spin_button_value_changed)
        self.evening_to_spin_button.connect("value-changed", self.day_part_spin_button_value_changed)

        # multi view spin button
        self.multi_view_spin_button = builder.get_object("multi_view_spin_button")
        preference_settings.bind("multi-view-num-days", self.multi_view_spin_button, "value", Gio.SettingsBindFlags.DEFAULT)
        self.multi_view_spin_button.connect("value-changed", self.on_multi_view_spin_button_value_changed)

        # multi view grouping events
        self.multi_view_grouping_combo = builder.get_object("multi_view_grouping_combo")
        preference_settings.bind("multi-view-num-grouping-events", self.multi_view_grouping_combo, "active-id", Gio.SettingsBindFlags.DEFAULT)
        self.multi_view_grouping_combo.connect("changed", self.on_multi_view_grouping_combo_value_changed)

        # scale thumbnail to center / fill radio buttons
        self.scale_to_center_radio_button = builder.get_object("scale_to_center_radiobutton")
        self.scale_to_fill_radio_button = builder.get_object("scale_to_fill_radiobutton")
        self.scale_to_center_radio_button.connect("toggled", self.scale_to_radio_button_toggled)
        if preference_settings["scale-to-center-thumbnail"]:
            self.scale_to_center_radio_button.set_active(True)
        else:
            self.scale_to_fill_radio_button.set_active(True)

        # show thumbnails in timeline check button
        self.show_thumbnails_in_timeline = builder.get_object("show_thumbnails_in_timeline_checkbutton")
        self.show_thumbnails_in_timeline.connect("toggled", self.show_thumbnails_in_timeline_toggled)
        preference_settings.bind("show-thumbnails-in-timeline", self.show_thumbnails_in_timeline, "active", Gio.SettingsBindFlags.DEFAULT)

        # show day / night gradient in timeline header check button
        self.show_day_night_gradient_in_timeline = builder.get_object("show_day_night_in_timeline_checkbutton")
        self.show_day_night_gradient_in_timeline.connect("toggled", self.show_day_night_in_timeline_toggled)
        preference_settings.bind("show-day-night-gradient-in-timeline", self.show_day_night_gradient_in_timeline, "active", Gio.SettingsBindFlags.DEFAULT)

    def show(self):
        self.preferences_dialog.set_transient_for(self.window)
        self.preferences_dialog.present()

    def separate_zoom_per_view_check_button_toggled(self, button):
        config.SEPARATE_ZOOM_PER_VIEW = button.get_active()

    def delete_confirm_check_button_toggled(self, button):
        config.CONFIRM_ITEM_DELETION = button.get_active()

    def hide_empty_day_labels_check_button_toggled(self, button):
        config.HIDE_EMPTY_DAY_LABELS = button.get_active()
        self.show_info_bar_restart_app()

    def enable_tray_icon_check_button_toggled(self, button):
        enabled = False
        # update gsettings manually
        if TrayIconManager.check_dependencies():
            enabled = button.get_active()
        else:
            label=_("Please install an appIndicator3 gir package.")
            self.show_info_bar(label)

        preference_settings["enable-tray-icon"] = enabled
        config.ENABLE_TRAY_ICON = enabled

    def show_thumbnails_in_timeline_toggled(self, button):
        config.SHOW_THUMBNAILS_IN_TIMELINE = button.get_active()

    def show_day_night_in_timeline_toggled(self, button):
        config.SHOW_DAY_NIGHT_GRADIENT_IN_TIMELINE = button.get_active()

    def scale_to_radio_button_toggled(self, button):
        state = button.get_active()
        preference_settings["scale-to-center-thumbnail"] = state
        config.SCALE_TO_CENTER = state

    def day_part_spin_button_value_changed(self, button):
        self.show_info_bar_restart_app()

    def on_multi_view_spin_button_value_changed(self, button):
        self.show_info_bar_restart_app()

    def on_multi_view_grouping_combo_value_changed(self, combo):
        self.show_info_bar_restart_app()

    def search_in_subject_uri_check_button_toggled(self, button):
        config.SEARCH_IN_SUBJECT_URI = button.get_active()

    def search_ignore_single_char_check_button_toggled(self, button):
        config.SEARCH_IGNORE_SINGLE_CHAR = button.get_active()

    def search_is_case_sensitive_check_button_toggled(self, button):
        config.SEARCH_IS_CASE_SENSITIVE = button.get_active()

    def search_results_use_bold_check_button_toggled(self, button):
        enabled = button.get_active()
        if not self.search_results_highlight_check():
            enabled = not enabled
        preference_settings["search-results-use-bold-style"] = enabled
        config.SEARCH_RESULTS_USE_BOLD_STYLE = enabled

    def search_results_use_large_check_button_toggled(self, button):
        enabled = button.get_active()
        if not self.search_results_highlight_check():
            enabled = not enabled
        preference_settings["search-results-use-large-style"] = enabled
        config.SEARCH_RESULTS_USE_LARGE_STYLE = enabled

    def search_results_use_default_color_check_button_toggled(self, button):
        enabled = button.get_active()
        if not self.search_results_highlight_check():
            enabled = not enabled
        preference_settings["search-results-use-default-color"] = enabled
        config.SEARCH_RESULTS_USE_DEFAULT_COLOR = enabled

        self.search_results_text_color_button.set_sensitive(not enabled)
        self.search_results_histogram_color_button.set_sensitive(not enabled)

    def search_results_text_color_changed(self, color_button):
        text_color = Gdk.RGBA.to_string(color_button.get_rgba())
        preference_settings["search-results-text-color"] = text_color
        preference_settings["search-results-text-color-updated"] = True
        config.SEARCH_RESULTS_TEXT_COLOR = text_color
        config.SEARCH_RESULTS_TEXT_COLOR_UPDATED = True

    def search_results_histogram_color_changed(self, color_button):
        histogram_color = Gdk.RGBA.to_string(color_button.get_rgba())
        preference_settings["search-results-histogram-color"] = histogram_color
        preference_settings["search-results-histogram-color-updated"] = True
        config.SEARCH_RESULTS_HISTOGRAM_COLOR = histogram_color
        config.SEARCH_RESULTS_HISTOGRAM_COLOR_UPDATED = True

    def show_audio_preview_check_button_toggled(self, button):
        enabled = False
        # update gsettings manually
        if AudioPreviewTooltip and AudioPreviewTooltip.check_dependencies():
            self.show_info_bar_restart_app()
            enabled = button.get_active()
        else:
            label=_("Please install gstreamer1.0 gir package.")
            self.show_info_bar(label)

        preference_settings["show-audio-preview"] = enabled

    def show_video_preview_check_button_toggled(self, button):
        enabled = False
        # update gsettings manually
        if VideoPreviewTooltip and VideoPreviewTooltip.check_dependencies():
            self.show_info_bar_restart_app()
            enabled = button.get_active()
        else:
            label=_("Please install gstreamer1.0-gtk3 package.")
            self.show_info_bar(label)

        preference_settings["show-video-preview"] = enabled

    def show_info_bar(self, label):
        self.info_bar_label.set_text(label)
        self.info_bar.set_revealed(True)
        self.info_bar.show_all()

    def show_info_bar_restart_app(self):
        label=_("Restart application for changes to take effect.")
        self.show_info_bar(label)

    def on_info_bar_response(self, info_bar, response):
        self.info_bar.hide()

    def search_results_highlight_check(self):
        bold = self.search_results_use_bold_check_button.get_active()
        large = self.search_results_use_large_check_button.get_active()
        default_color = self.search_results_use_default_color_check_button.get_active()

        if not (bold or large):
            if default_color or not config.SEARCH_RESULTS_TEXT_COLOR_UPDATED:
                label = _("Custom text color or bold or large style should be enabled ")
                self.show_info_bar(label)
                return False
        return True



# -*- Mode: python; coding: utf-8; tab-width: 4; indent-tabs-mode: nil; -*-
#
# GNOME Activity Journal
#
# Copyright © 2010 Randal Barlow
# Copyright © 2010 Siegfried Gevatter <siegfried@gevatter.com>
# Copyright © 2010 Stefano Candori <stefano.candori@gmail.com>
# Copyright © 2020 The GNOME Activity Journal developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime
import dbus
import sys
import time
import threading
from zeitgeist.datamodel import Event, ResultType, Interpretation, TimeRange, \
    Subject, StorageState

from . import content_objects
from . import external
from . import config
from .external import CLIENT, CLIENT_EXTENSION

import gi
from gi.repository import GObject, Gtk, Gdk

MAXEVENTS = 999999

tdelta = lambda x: datetime.timedelta(days=x)

def get_related_events_for_uri(uri, callback):
    """
    :param uri: A uri for which to request related uris using zeitgeist
    :param callback: this callback is called once the events are retrieved for
    the uris. It is called with a list of events.
    """
    def _event_request_handler(ids):
        events = []
        for id_ in ids:
            try:
                events.append(STORE.get_event_from_id(id_))
            except KeyError as e:
                print("%s" % e)
        callback(events)

    def _event_id_request_handler(uris):
        templates = []
        if len(uris) > 0:
            for i, uri in enumerate(uris):
                sub = Subject.new_for_values(uri=uri)
                templates += [
                        Event.new_for_values(subjects=[sub]),
                    ]
            CLIENT.find_event_ids_for_templates(templates, _event_request_handler,
                TimeRange.until_now(), num_events=len(uris),
                storage_state=StorageState.Available,
                result_type=ResultType.MostRecentSubjects)

    end = time.time() * 1000
    start = end - (86400*30*1000)
    CLIENT.find_related_uris_for_uris([uri], _event_id_request_handler)


def reduce_dates_by_timedelta(dates, delta):
    new_dates = []
    for date in dates:
        new_dates += [date + delta]
    return new_dates


class DoEmit(object):
    """
    Calls emit_update on the methods class
    """
    def __init__(self, signal):
        self.signal = signal

    def __call__(self, function):
        def wrapper(instance, *args, **kwargs):
            if (isinstance(instance, Day)):
                setattr(instance, "_loaded", True)
            value = function(instance, *args, **kwargs)
            instance.emit(self.signal)
            return value
        return wrapper


class CachedAttribute(object):
    """
    runs the method once, finds the value, and replace the descriptor
    in the instance with the found value
    """
    def __init__(self, method, name=None):
        self.method = method
        self.attr_name = name or method.__name__

    def __get__(self, instance, cls):
        if instance is None:
            return self
        value = self.method(instance)
        setattr(instance, self.attr_name, value)
        return value


class ContentStruct(object):
    id = 0
    event = None

    _content_object_built = False
    @CachedAttribute
    def content_object(self):
        obj = None
        self._content_object_built = True
        try:
            obj = content_objects.ContentObject.new_from_event(self.event)
            if not obj:
                print("Error: no content object type found for event: " + str(self.event))
        except Exception as e:
            if config.DEBUG:
                print("error: %s" % str(e))
        return obj

    @CachedAttribute
    def event(self):
        events = CLIENT._iface.GetEvents([self.id])
        if events:
            return Event(events[0])

    def __str__(self):
        return str(self.event.subjects[0].uri)

    def __init__(self, id, event=None, content_object=None, build=False):
        self.id = id
        if event:
            self.event = event
        if content_object:
            self.content_object = content_object
        if build:
            CLIENT.get_events([self.id], self.set_event)

    def set_event(self, value):
        if isinstance(value, dbus.Array) or isinstance(value, list) or isinstance(value, tuple) and len(value):
            self.event = value[0]
        elif isinstance(value, Event):
            self.event = value
        else:
            self.event = ContentStruct.event
        self.build_struct()

    def build_struct(self):
        Gdk.threads_enter()
        self._content_object_built = True
        if not self.event.subjects[0].uri.startswith("http"):
            self.content_object = content_objects.ContentObject.new_from_event(self.event)
        Gdk.threads_leave()

class Day(GObject.GObject):
    __gsignals__ = {
        "update" : (GObject.SignalFlags.RUN_LAST, None, ())
    }


    date = None
    start = 0
    end = 86400

    @property
    def time_range(self):
        return [self.start*1000, self.end*1000]

    @property
    def items(self):
        self.load_ids()
        return list(self._items.values())

    @CachedAttribute
    def templates(self):
        subject = Subject()
        subject.uri = "!application://*"
        return[Event.new_for_values(subjects = [subject], actor="!application://activity-log-manager.desktop")]

    def __str__(self):
        return time.asctime(time.localtime(self.start))

    def __init__(self, date, days_population=None):
        super(Day, self).__init__()
        self.date = date
        self._items = {}#id:ContentItem
        self._load_requested = False
        self._loaded = False
        self.start = int(time.mktime(date.timetuple()))
        self.end = self.start+86399
        self._population = None
        if days_population:
            for timestamp, count in days_population: # they are ordered descending
                # Ugly hack to adjust for local/UTC time. Screw you timezones!
                timestamp += (time.gmtime().tm_hour - time.localtime().tm_hour) * 3600 + \
                    (time.gmtime().tm_min - time.localtime().tm_min) * 60
                if timestamp >= self.start:
                    if timestamp < self.end:
                        self._population = count
                        break
                else:
                    break
        else:
            self.load_ids()
        if external.HAMSTER:
            try:
                facts = external.HAMSTER.get_facts(self.start, self.end)
                for fact in facts:
                    self.insert_events(None, fact.get_events())
            except TypeError:
                pass #print "Hamster support disabled temporarely"

    def is_loaded(self):
        return self._loaded

    def load_ids(self):
        if not self._load_requested:
            self._load_requested = True
            CLIENT.find_events_for_templates(self.templates, self.set_ids,
                self.time_range, num_events=MAXEVENTS,
                storage_state=StorageState.Available)
            CLIENT.install_monitor(self.time_range, self.templates, self.insert_events, self.remove_ids)

    def __getitem__(self, id_):
        self.load_ids()
        return self._items[id_]

    def __len__(self):
        if self._load_requested:
            return len(self._items)
        return self._population or 0

    def has_id(self, id_):
        if not self._load_requested: self.load_ids()
        return id_ in self._items

    @DoEmit("update")
    def set_ids(self, event_ids):
        deleted_uris = STORE.list_deleted_uris
        for event in event_ids:
            #let's update the GUI
            Gdk.threads_enter()
            while Gtk.events_pending():Gtk.main_iteration()
            Gdk.threads_leave()
            if deleted_uris is not None:
                if event.subjects[0].uri not in deleted_uris:
                    self._items[event.id] = ContentStruct(event.id, event)
            else:
                self._items[event.id] = ContentStruct(event.id, event)

    @DoEmit("update")
    def remove_ids(self, time_range, ids):
        for id_ in ids:
            try:
                del self._items[id_]
            except KeyError:
                pass

    @DoEmit("update")
    def insert_events(self, time_range, events):
        for event in events:
            self._items[event.id] = ContentStruct(event.id, event)

    @DoEmit("update")
    def insert_event(self, event, overwrite=False):
        """
        Insert an event into the day object

        Emits a 'update' signal
        """
        return self._insert_event(event, overwrite)

    def _insert_event(self, event, overwrite=False):
        """
        Insert an event into the day object without emitting a 'update' signal
        """
        if not overwrite and event.id in self._items:
            self._items[event.id].event = event
            return False
        self._items[event.id] = ContentStruct(event.id, event)
        if overwrite:
            self._items[event.id]._content_object_built = True
            self._items[event.id].content_object = content_objects.ContentObject.new_from_event(event)
        return True

    def next(self, store=None):
        """
        Return the next day in the given store
        """
        if not store:
            store = STORE # Singleton
        date = self.date + datetime.timedelta(days=1)
        return store[date]

    def previous(self, store=None):
        """
        Return the previous day in the given store
        """
        if not store:
            store = STORE # Singleton
        date = self.date + datetime.timedelta(days=-1)
        return store[date]

    def filter(self, event_template=None, result_type=None):
        self.load_ids()
        if event_template:
            items = self.filter_event_template(event_template)
            items = self.filter_result_type(items, result_type)
        elif result_type:
            items = self.filter_result_type(list(self._items.values()), result_type)
        else:
            items = list(self._items.values())
        #I reverse the list to make MODIFY/ACCESS_EVENT "more important" than CREATE ones
        #Doing that, fx. tomboy's note names are updated - cando
        items.sort(key=lambda obj: int(obj.event.timestamp),reverse=True)
        return items

    def filter_event_template(self, event_template):
        items = []
        if isinstance(event_template, Event):
            for obj in list(self._items.values()):
                if obj.event.matches_template(event_template):
                    items.append(obj)
        elif event_template:
            for template in event_template:
                items += self.filter(template)
            items = list(set(items))
        return items

    def filter_result_type(self, items, result_type):
        if items and result_type is ResultType.MostRecentSubjects:
            item_dict = {}
            for item in items:
                subject_uri = item.event.subjects[0].uri
                item_dict[subject_uri] = item
            items = list(item_dict.values())
        return items

    def get_time_map(self):
        start = self.start
        results = {}
        for item in self.items:
            uri = item.event.subjects[0].uri
            if uri.startswith("http://") or uri.startswith("https://"): continue
            if not uri in results:
                results[uri] = []
            if not item.event.interpretation == Interpretation.LEAVE_EVENT.uri:
                results[uri].append([item, 0])
            else:
                if not len(results[uri]) == 0:
                    #print "***", results[uri]
                    results[uri][len(results[uri])-1][1] = (int(item.event.timestamp)) -  int(results[uri][-1][0].event.timestamp)
                else:
                    tend = int(item.event.timestamp)/1000
                    item.event.timestamp = str(start)
                    results[uri].append([item, tend - start])
        results = list(sorted(iter(results.values()), key=lambda r: \
                             r[0][0].event.timestamp))
        return results

    def __set_events_by_id(self, events):
        for event in events:
            self._items[event.id].event = event
            self._items[event.id].build_struct()


class Store(GObject.GObject):
    __gsignals__ = {
        "update" : (GObject.SignalFlags.RUN_LAST, None, ())
    }

    @property
    def today(self):
        return self[datetime.date.today()]

    @property
    def days(self):
        dates = self.dates
        return [self._days[date] for date in dates]

    @property
    def dates(self):
        dates = list(self._days.keys())
        dates.sort()
        return dates

    @property
    def list_deleted_uris(self):
        return self._deleted_uris

    @property
    def loaded_items(self):
        for day in self.days:
            for item in day.items:
                yield item

    def __init__(self):
        super(Store, self).__init__()
        self.run_build_thread = False
        self._days = {}
        self._day_connections = {}
        self._deleted_uris = []
        #Search for uris that have been deleted in order to not display them.
        #FIXME we should add a timestamp field with the deleted uri
        #to prevent that a recent event with same uri than an older and deleted one
        #isn't displayed. - cando
        self._deleted_uris = []
        subject = Subject()
        subject.uri = "!application://*"
        template = Event.new_for_values(interpretation=Interpretation.DELETE_EVENT.uri,
                                        subjects = [subject],
                                        actor="!application://activity-log-manager.desktop")

        CLIENT.find_events_for_templates((template,), self.__set_deleted_uris,
            TimeRange.until_now(), num_events=MAXEVENTS)
        global currentTimestamp, histogramLoaderCounter
        today = datetime.date.today()
        currentTimestamp = time.mktime(today.timetuple())
        days_population = CLIENT_EXTENSION.GetHistogramData()
        for i in range(50 * 6):
            date = datetime.date.fromtimestamp(currentTimestamp)
            day = Day(date, days_population)
            self.add_day(date, day)
            currentTimestamp -= 86400
        for day in self.days:
            day.load_ids()
        content_objects.AbstractContentObject.connect_to_manager("add", self.add_content_object_with_new_type)
        content_objects.AbstractContentObject.connect_to_manager("remove", self.remove_content_objects_with_type)

    def __set_deleted_uris(self, ids):
        for event in ids:
            self._deleted_uris.append(event.subjects[0].uri)
        return self._deleted_uris

    def add_content_object_with_new_type(self, obj):
        for day in self.days:
            for instance in day.items:
                if instance._content_object_built:
                    cls = content_objects.ContentObject.find_best_type_from_event(instance.event)
                    if not isinstance(instance.content_object, cls) and instance.content_object:
                        del instance.content_object
                        instance.content_object = cls.create(instance.event)
            day.emit("update")

    def remove_content_objects_with_type(self, obj):
        for day in self.days:
            for instance in day.items:
                if instance._content_object_built:
                    if isinstance(instance.content_object, obj) and instance.content_object:
                        instance.content_object = content_objects.ContentObject.new_from_event(instance.event)
            day.emit("update")

    @DoEmit("update")
    def add_day(self, key, day):
        self._days[key] = day
        self._day_connections[key] = day.connect("update", lambda *args: self.emit("update"))

    def get_event_from_id(self, id_):
        struct = ContentStruct(id_)
        date = datetime.date.fromtimestamp(int(struct.event.timestamp)/1000)
        if date in self.dates:
            nstruct = self[date][id_]
            nstruct.event = struct.event
            del struct
            return nstruct
        else:
            day = Day(date)
            self.add_day(date, day)
            day._items[id_] = struct
            return struct

    def __getitem__(self, key):
        if isinstance(key, datetime.date):
            # Return day from date
            try:
                return self._days[key]
            except KeyError:
                day = Day(key)
                self.add_day(key, day)
                return day
        elif isinstance(key, int):
            # Return event id
            for date, obj in self._days.items():
                if obj.has_id(id):
                    return obj[id]
        raise KeyError("%s Not found" % key)

    def __len__(self):
        i = 0
        for item in self._days.values():
            i+=len(item)
        return i

    def request_last_n_days_events(self, n=90, func=None):
        """
        Find the events for 'n' days and packs them into the store

        Optionally calls func upon completion
        """
        subject = Subject()
        subject.uri = "!application://*"
        event_templates = (
            Event.new_for_values(interpretation=Interpretation.ACCESS_EVENT.uri, subjects=[subject]),
            Event.new_for_values(interpretation=Interpretation.MODIFY_EVENT.uri, subjects=[subject]),
            Event.new_for_values(interpretation=Interpretation.CREATE_EVENT.uri, subjects=[subject]),
            Event.new_for_values(interpretation=Interpretation.RECEIVE_EVENT.uri, subjects=[subject]),
            Event.new_for_values(actor="!application://activity-log-manager.desktop")
        )


        def callback(events):
            def _thread_fn(events):
                event_chunks = []
                chunk = []
                for i, event in enumerate(events):
                    chunk.append(event)
                    if i%20 == 0:
                        event_chunks.append(chunk)
                        chunk = []
                nolazy = list(map(self.add_events, event_chunks))
                if func:
                    func()
                return False

            thread = threading.Thread(target=_thread_fn, args=(events,))
            thread.start()

        end = time.time() - 3*86400
        start = end - n*86400
        if n >= 60:
            inc = (end - start)/(n/30)
            for i in range(n/30):
                a = end - (inc*(i+1)) + 1
                b = end - (inc*(i))
                CLIENT.find_events_for_templates(
                    event_templates, callback, [a*1000, b*1000],
                    num_events=50000, storage_state=StorageState.Available)
        else:
            CLIENT.find_events_for_templates(
                event_templates, callback, [start*1000, end*1000],
                num_events=50000, storage_state=StorageState.Available)
        return False

    def __add_event(self, event, overwrite):
        date = datetime.date.fromtimestamp(int(event.timestamp)/1000)
        day = self[date]
        day._insert_event(event, overwrite)

    def __add_events(self, events, overwrite):
        for event in events:
            self.__add_event(event, overwrite)

    def add_events(self, events, overwrite=True, idle=True):
        if idle:
            def _idle_add(events, overwrite):
                # Use _insert_event to avoid update signals
                self.__add_events(events, overwrite)
                return False
            GObject.idle_add(_idle_add, events, overwrite)
        else:
            self.__add_events(events, overwrite)

    def search_store_using_matching_function(self, func, date=None):
        matches = []
        items = self.loaded_items if not date else self[date].items
        for item in items:
            if func(item):
                matches.append(item)
        return matches

    def search_using_zeitgeist_fts(self, text, event_templates=None):
        if not external.FTS:
            return []
        events = external.FTS.search(text, event_templates if event_templates else [])
        ids = [int(event.id) for event in events]
        results = []
        for id_ in ids:
            try:
                event = self.get_event_from_id(id_)
                results.append(event)
            except KeyError as e:
                pass
        return results

    @property
    def fts_search_enabled(self):
        # Disabled FTS search until it is further refined
        if external.FTS: return True
        return False


GObject.type_register(Day)
GObject.type_register(Store)

# Init
STORE = Store()
external.STORE = STORE



# -*- Mode: python; coding: utf-8; tab-width: 4; indent-tabs-mode: nil; -*-
#
# GNOME Activity Journal
#
# Copyright © 2009-2010 Seif Lotfy <seif@lotfy.com>
# Copyright © 2009-2010 Siegfried Gevatter <siegfried@gevatter.com>
# Copyright © 2007 Alex Graveley <alex@beatniksoftware.com>
# Copyright © 2010 Markus Korn <thekorn@gmx.de>
# Copyright © 2010 Randal Barlow <email.tehk@gmail.com>
# Copyright © 2020 The GNOME Activity Journal developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import gettext

from zeitgeist.datamodel import Interpretation

# Legacy issues
INTERPRETATION_UNKNOWN = "" # "http://zeitgeist-project.com/schema/1.0/core#UnknownInterpretation"
MANIFESTATION_UNKNOWN = "" # "http://zeitgeist-project.com/schema/1.0/core#UnknownManifestation"
INTERPRETATION_NOTE = "aj://note"
INTERPRETATION_VCS = "aj://vcs" #version control system

class Source(object):
    """A source which is used for category creation and verb/description storage for a given interpretation"""
    def __init__(self, interpretation, icon, desc_sing, desc_pl):
        if not isinstance(interpretation, str):
            self.name = interpretation.name
        else:
            self.name = interpretation
        self.icon = icon
        self._desc_sing = desc_sing
        self._desc_pl = desc_pl

    def group_label(self, num):
        if num == 1: return gettext.gettext(self._desc_sing)
        else: return gettext.gettext(self._desc_pl)

SUPPORTED_SOURCES = {
    # TODO: Move this into Zeitgeist's library, implemented properly
    Interpretation.VIDEO.uri: Source(Interpretation.VIDEO, "gnome-mime-video", _("Worked with a Video"), _("Worked with Videos")),
    Interpretation.AUDIO.uri: Source(Interpretation.AUDIO, "gnome-mime-audio", _("Worked with Audio"), _("Worked with Audio")),
    Interpretation.IMAGE.uri: Source(Interpretation.IMAGE, "image", _("Worked with an Image"), _("Worked with Images")),
    Interpretation.DOCUMENT.uri: Source(Interpretation.DOCUMENT, "gnome-word", _("Edited or Read Document"), _("Edited or Read Documents")),
    Interpretation.SOURCE_CODE.uri: Source(Interpretation.SOURCE_CODE, "applications-development", _("Edited or Read Code"), _("Edited or Read Code")),
    Interpretation.IMMESSAGE.uri: Source(Interpretation.IMMESSAGE, "applications-internet", _("Conversation"), _("Conversations")),
    Interpretation.WEBSITE.uri: Source(Interpretation.WEBSITE, "gnome-web-browser", _("Visited Website"), _("Visited Websites")),
    Interpretation.EMAIL.uri: Source(Interpretation.EMAIL, "applications-internet", _("Email"), _("Emails")),
    Interpretation.TODO.uri: Source(Interpretation.TODO, "applications-office", _("Todo"), _("Todos")),
    INTERPRETATION_UNKNOWN: Source("Unknown", "applications-other", _("Other Activity"), _("Other Activities")),
    INTERPRETATION_NOTE: Source("aj://note", "tomboy", _("Edited or Read Note"), _("Edited or Read Notes")),
    INTERPRETATION_VCS: Source("aj://vcs", "bzr-icon-64", _("Software Development"), _("Software Developments"))
}



# -*- Mode: python; coding: utf-8; tab-width: 4; indent-tabs-mode: nil; -*-
#
# History Manager
#
# Copyright © 2011 Collabora Ltd.
#             By Siegfried-Angel Gevatter Pujals <siegfried@gevatter.com>
# Copyright © 2020 The GNOME Activity Journal developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.



from zeitgeist.datamodel import Event, Subject

from .external import CLIENT, CLIENT_VERSION

class BlacklistInterface:

    INCOGNITO = Event.new_for_values()
    INCOGNITO_ID = 'block-all'

    _callback = None

    def __init__(self):
        self._blacklist = CLIENT._iface.get_extension('Blacklist', 'blacklist')
        self._blacklist.connect('TemplateAdded', self._on_notification)
        self._blacklist.connect('TemplateRemoved', self._on_notification)

    def set_incognito_toggle_callback(self, callback):
        self._callback = callback

    def _on_notification(self, blacklist_id, template):
        if self.INCOGNITO.matches_template(template):
            if self._callback is not None:
                self._callback()

    def _get_blacklist_templates(self):
        return self._blacklist.GetTemplates()

    def _add_blacklist_template(self, template):
        self._blacklist.AddTemplate(self.INCOGNITO_ID, self.INCOGNITO)

    def _remove_blacklist_template(self, template_id):
        # FIXME: This should be changed to something that doesn't depend on
        # INCOGNITO_ID, but just on the INCOGNITO template.
        self._blacklist.RemoveTemplate(self.INCOGNITO_ID)
        self._blacklist.RemoveTemplate('incognito') # for backwards compatibility

    def get_incognito(self):
        templates = self._get_blacklist_templates()
        return any(map(self.INCOGNITO.matches_template, iter(templates.values())))

    def toggle_incognito(self):
        if not self.get_incognito():
            self._add_blacklist_template(self.INCOGNITO)
        else:
            self._remove_blacklist_template(self.INCOGNITO)

BLACKLIST = BlacklistInterface()



# -*- Mode: python; coding: utf-8; tab-width: 4; indent-tabs-mode: nil; -*-
#
# GNOME Activity Journal
#
# Copyright © 2010 Randal Barlow <email.tehk@gmail.com>
# Copyright © 2010 Siegfried Gevatter <siegfried@gevatter.com>
# Copyright © 2010 Markus Korn <thekorn@gmx.de>
# Copyright © 2020 The GNOME Activity Journal developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Common Functions and classes which are used to create the alternative views,
and handle colors, pixbufs, and text
"""


import collections
import gettext
import os
import time
import math
import operator
import subprocess
import tempfile
import urllib.request, urllib.parse, urllib.error
import zipfile
import gi
import cairo
import sys

gi.require_version('PangoCairo', '1.0')

try:
    gi.require_version('GnomeDesktop', '3.0')
except ValueError as e:
    print("ERROR: %s." % e)
    print()
    print("Please install 'gnomedesktop-3.0' gir package.")
    print()
    print("See the README.md file for more information.")
    sys.exit(1)

from gi.repository import GObject, Gtk, GLib, Gdk, Gio, GnomeDesktop
from gi.repository import Pango, PangoCairo, GdkPixbuf

try:
    import pygments
except ImportError:
    pygments = None
else:
    from pygments.lexers import get_lexer_for_filename, get_lexer_for_mimetype
    from pygments import highlight
    from pygments.formatters import ImageFormatter

try:
    import chardet
except ImportError:
    chardet = None

from . import config
from .config import get_data_path, get_icon_path
from .sources import INTERPRETATION_UNKNOWN, MANIFESTATION_UNKNOWN

from zeitgeist.datamodel import Interpretation, Event, Subject


THUMBS = collections.defaultdict(dict)
ICONS = collections.defaultdict(dict)
ICON_SIZES = SIZE_NORMAL, SIZE_LARGE = ((128, 128), (256, 256))

# Thumbview and Timelineview sizes
SIZE_THUMBVIEW = [(48+32, 36+32), (96+32, 72+32), (150+32, 108+32), (256+32, 192+32)]
SIZE_TIMELINEVIEW = [(16, 12), (32, 24), (64, 48), (96, 72)]
SIZE_TEXT_TIMELINEVIEW = ["small", "medium", "large", "large"]
SIZE_TEXT_THUMBVIEW = ["x-small", "small", "medium", "large"]

APP_ICON_NORMAL = "hicolor/scalable/apps/org.gnome.ActivityJournal.svg"
APP_ICON_PAUSED = "hicolor/scalable/apps/org.gnome.ActivityJournal-paused.svg"

THUMBNAIL_FACTORIES = {
    SIZE_NORMAL: GnomeDesktop.DesktopThumbnailFactory.new(GnomeDesktop.DesktopThumbnailSize.NORMAL),
    SIZE_LARGE: GnomeDesktop.DesktopThumbnailFactory.new(GnomeDesktop.DesktopThumbnailSize.LARGE)
}
ICON_THEME = Gtk.IconTheme.get_default()

# Caches desktop files
DESKTOP_FILES = {}
DESKTOP_FILE_PATHS = []
try:
    desktop_file_paths = os.environ["XDG_DATA_DIRS"].split(":")
    for path in desktop_file_paths:
        if path.endswith("/"):
            DESKTOP_FILE_PATHS.append(path + "applications/")
        else:
            DESKTOP_FILE_PATHS.append(path + "/applications/")
except KeyError:pass

# Placeholder pixbufs for common sizes
PLACEHOLDER_PIXBUFFS = {
    24 : GdkPixbuf.Pixbuf.new_from_file_at_size(get_icon_path(APP_ICON_NORMAL), 24, 24),
    16 : GdkPixbuf.Pixbuf.new_from_file_at_size(get_icon_path(APP_ICON_NORMAL), 16, 16)
    }

# Color magic
TANGOCOLORS = [
    (252/255.0, 234/255.0,  79/255.0),#0
    (237/255.0, 212/255.0,   0/255.0),
    (196/255.0, 160/255.0,   0/255.0),

    (252/255.0, 175/255.0,  62/255.0),#3
    (245/255.0, 121/255.0,   0/255.0),
    (206/255.0,  92/255.0,   0/255.0),

    (233/255.0, 185/255.0, 110/255.0),#6
    (193/255.0, 125/255.0,  17/255.0),
    (143/255.0,  89/255.0,   2/255.0),

    (138/255.0, 226/255.0,  52/255.0),#9
    (115/255.0, 210/255.0,  22/255.0),
    ( 78/255.0, 154/255.0,   6/255.0),

    (114/255.0, 159/255.0, 207/255.0),#12
    ( 52/255.0, 101/255.0, 164/255.0),
    ( 32/255.0,  74/255.0, 135/255.0),

    (173/255.0, 127/255.0, 168/255.0),#15
    (117/255.0,  80/255.0, 123/255.0),
    ( 92/255.0,  53/255.0, 102/255.0),

    (239/255.0,  41/255.0,  41/255.0),#18
    (204/255.0,   0/255.0,   0/255.0),
    (164/255.0,   0/255.0,   0/255.0),

    (136/255.0, 138/255.0, 133/255.0),#21
    ( 85/255.0,  87/255.0,  83/255.0),
    ( 46/255.0,  52/255.0,  54/255.0),
]

FILETYPES = {
    Interpretation.VIDEO.uri : 0,
    Interpretation.AUDIO.uri : 3,
    Interpretation.DOCUMENT.uri : 12,
    Interpretation.IMAGE.uri : 15,
    Interpretation.SOURCE_CODE.uri : 12,
    INTERPRETATION_UNKNOWN : 21,
    Interpretation.IMMESSAGE.uri : 21,
    Interpretation.EMAIL.uri : 21
}

FILETYPESNAMES = {
    Interpretation.VIDEO.uri : _("Video"),
    Interpretation.AUDIO.uri : _("Music"),
    Interpretation.DOCUMENT.uri : _("Document"),
    Interpretation.IMAGE.uri : _("Image"),
    Interpretation.SOURCE_CODE.uri : _("Source Code"),
    INTERPRETATION_UNKNOWN : _("Unknown"),
    Interpretation.IMMESSAGE.uri : _("IM Message"),
    Interpretation.EMAIL.uri :_("Email"),

}

INTERPRETATION_PARENTS = {
    Interpretation.DOCUMENT.uri : Interpretation.DOCUMENT.uri,
    Interpretation.TEXT_DOCUMENT.uri : Interpretation.DOCUMENT.uri,
    Interpretation.PAGINATED_TEXT_DOCUMENT.uri : Interpretation.DOCUMENT.uri,
    Interpretation.RASTER_IMAGE.uri : Interpretation.IMAGE.uri,
    Interpretation.VECTOR_IMAGE.uri : Interpretation.IMAGE.uri,
    Interpretation.IMAGE.uri : Interpretation.IMAGE.uri,
    Interpretation.AUDIO.uri : Interpretation.AUDIO.uri,
    Interpretation.MUSIC_PIECE.uri : Interpretation.MUSIC_PIECE.uri,
}

MEDIAINTERPRETATIONS = [
    Interpretation.VIDEO.uri,
    Interpretation.IMAGE.uri,
]

def get_hex_color(color):
    return ("#%x%x%x" % (int(color.red * 255), int(color.green * 255), int(color.blue * 255)))

def get_file_color(ftype, fmime):
    """Uses hashing to choose a shade from a hue in the color tuple above

    :param ftype: a :class:`Event <zeitgeist.datamodel.Interpretation>`
    :param fmime: a mime type string
    """
    if ftype in list(FILETYPES.keys()):
        i = FILETYPES[ftype]
        l = int(math.fabs(hash(fmime))) % 3
        return TANGOCOLORS[min(i+l, len(TANGOCOLORS)-1)]
    return (136/255.0, 138/255.0, 133/255.0)

def get_text_or_uri(obj, ellipsize=True, molteplicity=False):
    if molteplicity and obj.molteplicity:
        text = obj.molteplicity_text
    else:
        text = str(obj.text.replace("&", "&amp;"))
    if text.strip() == "":
        text = str(obj.uri.replace("&", "&amp;"))
        text = urllib.parse.unquote(text)
    if len(text) > 50 and ellipsize: text = text[:50] + "..." #ellipsize--it's ugly!
    return text

##
## Zeitgeist event helper functions

def get_event_typename(event):
    """
    :param event: a :class:`Event <zeitgeist.datamodel.Event>`

    :returns: a plain text version of a interpretation
    """
    try:
        return Interpretation[event.subjects[0].interpretation].display_name
    except KeyError:
        pass
    return FILETYPESNAMES[event.subjects[0].interpretation]

##
# Cairo drawing functions

def draw_frame(context, x, y, w, h):
    """
    Draws a 2 pixel frame around a area defined by x, y, w, h using a cairo context

    :param context: a cairo context
    :param x: x position of the frame
    :param y: y position of the frame
    :param w: width of the frame
    :param h: height of the frame
    """
    x, y = int(x)+0.5, int(y)+0.5
    w, h = int(w), int(h)
    context.set_line_width(1)
    context.rectangle(x-1, y-1, w+2, h+2)
    context.set_source_rgba(0.5, 0.5, 0.5)#0.3, 0.3, 0.3)
    context.stroke()

    return

    # FixMe: The above frame looks good enough, and we don't the
    # extras below. So, we're returning there for now. Revisit this, if
    # necessary.
    context.set_source_rgba(0.7, 0.7, 0.7)
    context.rectangle(x, y, w, h)
    context.stroke()
    context.set_source_rgba(0.4, 0.4, 0.4)
    context.rectangle(x+1, y+1, w-2, h-2)
    context.stroke()

def draw_rounded_rectangle(context, x, y, w, h, r=5):
    """Draws a rounded rectangle

    :param context: a cairo context
    :param x: x position of the rectangle
    :param y: y position of the rectangle
    :param w: width of the rectangle
    :param h: height of the rectangle
    :param r: radius of the rectangle
    """
    context.new_sub_path()
    context.arc(r+x, r+y, r, math.pi, 3 * math.pi /2)
    context.arc(w-r+x, r+y, r, 3 * math.pi / 2, 0)
    context.arc(w-r+x, h-r+y, r, 0, math.pi/2)
    context.arc(r+x, h-r+y, r, math.pi/2, math.pi)
    context.close_path()
    return context

def draw_speech_bubble(context, layout, x, y, w, h):
    """
    Draw a speech bubble at a position

    Arguments:
    :param context: a cairo context
    :param layout: a pango layout
    :param x: x position of the bubble
    :param y: y position of the bubble
    :param w: width of the bubble
    :param h: height of the bubble
    """
    layout.set_width((w-10)*1024)
    layout.set_ellipsize(Pango.EllipsizeMode.MIDDLE)
    textw, texth = layout.get_pixel_size()
    context.new_path()
    context.move_to(x + 0.45*w, y+h*0.1 + 2)
    context.line_to(x + 0.5*w, y)
    context.line_to(x + 0.55*w, y+h*0.1 + 2)
    h = max(texth + 5, h)
    draw_rounded_rectangle(context, x, y+h*0.1, w, h, r = 5)
    context.close_path()
    context.set_line_width(2)
    context.set_source_rgb(168/255.0, 165/255.0, 134/255.0)
    context.stroke_preserve()
    context.set_source_rgb(253/255.0, 248/255.0, 202/255.0)
    context.fill()
    context.set_source_rgb(0, 0, 0)
    context.move_to(x+5, y+5)
    PangoCairo.update_layout(context, layout)
    PangoCairo.show_layout(context, layout)

def draw_text(context, layout, markup, x, y, maxw = 0, color = (0.3, 0.3, 0.3)):
    """
    Draw text using a cairo context and a pango layout

    Arguments:
    :param context: a cairo context
    :param layout: a pango layout
    :param x: x position of the bubble
    :param y: y position of the bubble
    :param maxw: the max text width in pixels
    :param color: a rgb tuple
    """
    layout.set_markup(markup)
    layout.set_ellipsize(Pango.EllipsizeMode.END)
    context.set_source_rgba(*color)
    if maxw:
        layout.set_width(maxw*1024)
    context.move_to(x, y)
    PangoCairo.update_layout(context, layout)
    PangoCairo.show_layout(context, layout)
    context.stroke()

def render_pixbuf(context, x, y, pixbuf, w, h, drawframe = True):
    """
    Renders a pixbuf to be displayed on the cell

    Arguments:
    :param context: a cairo context
    :param x: x position
    :param y: y position
    :param drawframe: if true we draw a frame around the pixbuf
    """
    context.save()
    context.rectangle(x, y, w, h)
    if drawframe:
        context.set_source_rgb(1, 1, 1)
        context.fill_preserve()
    Gdk.cairo_set_source_pixbuf (context, pixbuf, x, y)
    context.clip()
    context.paint()
    context.restore();
    if drawframe: # Draw a pretty frame
        draw_frame(context, x, y, w, h)

def render_emblems(context, x, y, w, h, emblems, emblem_size):
    """
    Renders emblems on the four corners of the rectangle

    Arguments:
    :param context: a cairo context
    :param x: x position
    :param y: y position
    :param w: the width of the rectangle
    :param y: the height of the rectangle
    :param emblems: a list of pixbufs
    """
    # w = max(self.width, w)
    corners = [[x, y],
               [x+w, y],
               [x, y+h],
               [x+w, y+h]]
    for i in range(len(emblems)):
        i = i % len(emblems)
        pixbuf = emblems[i]

        if pixbuf:
            # FixMe: Optimize
            if pixbuf.get_width() != emblem_size:
                pixbuf = pixbuf.scale_simple(emblem_size, emblem_size, GdkPixbuf.InterpType.BILINEAR)

            pbw, pbh = pixbuf.get_width(), pixbuf.get_height()
            pixbuf_x = corners[i][0] - pbw/2
            pixbuf_y = corners[i][1] - pbh/2

            Gdk.cairo_set_source_pixbuf(context, pixbuf, pixbuf_x, pixbuf_y)
            context.rectangle(pixbuf_x, pixbuf_y, pbw, pbh)
            context.fill()

def render_molteplicity(context, x, y, w, h, molteplicity, style, radius):
    """
    Renders the molteplicity number on the top-right corner of the rectangle

    Arguments:
    :param context: a cairo context
    :param x: x position
    :param y: y position
    :param w: the width of the rectangle
    :param y: the height of the rectangle
    :param emblems: the molteplicity number
    """
    center = [x + w, y ]

    color = get_gtk_rgba(style, "bg", 2)
    text_color = Gdk.RGBA(1, 1, 1, 1)

    # for border
    radius -= 1

    context.set_source_rgba(*color)
    context.arc(center[0], center[1], radius , 0, 2 * math.pi)
    context.fill_preserve()
    context.set_source_rgba(*color)
    context.stroke()

    context.set_source_rgba(*text_color)
    context.select_font_face("Cantarell", cairo.FontSlant.NORMAL,
            cairo.FontWeight.BOLD)

    if molteplicity < 10:
        font_size = radius
    elif molteplicity < 100:
        font_size = radius - 2
    else:
        font_size = radius - 3

    context.set_font_size(font_size)

    if molteplicity >= 100:
        text = "99+"
    else:
        text = str(molteplicity)

    x_bearing, y_bearing, width, height = context.text_extents(text)[:4]
    context.move_to(center[0] - width / 2 - x_bearing, center[1] - height / 2 - y_bearing)
    context.show_text(text)


##
## Color functions

def get_gdk_rgba_from_string(color_string):
    rgba = Gdk.RGBA()
    Gdk.RGBA.parse(rgba, color_string)

    return rgba

def shade_gdk_color(color, shade):
    """
    Shades a color by a fraction

    Arguments:
    :param color: a gdk color
    :param shade: fraction by which to shade the color

    :returns: a :class:`Color <Gdk.Color>`
    """
    f = lambda num: min((num * shade, 1.0))

    color.red = f(color.red)
    color.green = f(color.green)
    color.blue = f(color.blue)

    return color

def combine_gdk_color(color, fcolor):
    """
    Combines a color with another color

    Arguments:
    :param color: a gdk color
    :param fcolor: a gdk color to combine with color

    :returns: a :class:`Color <Gdk.Color>`
    """
    color.red = (2*color.red + fcolor.red)/3
    color.green = (2*color.green + fcolor.green)/3
    color.blue = (2*color.blue + fcolor.blue)/3

    return color

def get_histogram_default_search_color(style):
    pal = get_gtk_rgba(style, "bg", 3, 1.2)
    return Gdk.RGBA(pal.blue, pal.green, pal.red, 1)

def get_histogram_search_color(style):
    if config.SEARCH_RESULTS_USE_DEFAULT_COLOR:
        color = get_histogram_default_search_color(style)
    else:
        color = get_gdk_rgba_from_string(config.SEARCH_RESULTS_HISTOGRAM_COLOR)

    return color

def get_color_by_name(style, name):
    (found, bg) = style.lookup_color(name)

    if found:
        return bg

    raise Exception ("Color '%s' not found" % name)

def get_gtk_rgba(style, palette, i, shade = 1, alpha = 1):
    """Takes a gtk style and returns a RGB tuple

    Arguments:
    :param style: a gtk_style object
    :param palette: a string representing the palette you want to pull a color from
        Example: "bg", "fg"
    :param shade: how much you want to shade the color

    :returns: a rgba tuple
    """
    f = lambda num: (num) * shade

    color = None

    # FixMe: move to external css file
    if palette == "bg":
        if i == 0: # normal
            color = get_color_by_name(style, "theme_bg_color")
        elif i == 1: # prelight
            color = get_color_by_name(style, "theme_bg_color")
        elif i == 2: # selected
            color = get_color_by_name(style, "theme_selected_bg_color")
        elif i == 3: # insensitive
            color = get_color_by_name(style, "theme_selected_bg_color")
    elif palette == "fg":
        if i == 0:
            color = get_color_by_name(style, "theme_fg_color")
    elif palette == "text":
        if i == 0:
            color = get_color_by_name(style, "theme_text_color")
        elif i == 2: # selected
            color = get_color_by_name(style, "theme_text_color")
        elif i == 3: # insensitive
            color = get_color_by_name(style, "insensitive_base_color")
        elif i == 4: # active
            color = get_color_by_name(style, "theme_bg_color")
    elif palette == "base":
        if i == 0:
            color = get_color_by_name(style, "theme_base_color")
        elif i == 2: # selected
            color = get_color_by_name(style, "theme_selected_bg_color")
    else:
        raise Exception("Unexpected pallete value: %s" % (palette))

    if not color:
        raise Exception("Missing color for specified palette and index")

    if isinstance(color, Gdk.RGBA):
        red = f(color.red)
        green = f(color.green)
        blue = f(color.blue)

        return Gdk.RGBA(red=min(red, 1), green=min(green, 1), blue=min(blue, 1), alpha=alpha)
    else:
        raise TypeError("Not a valid Gdk.RGBA color")


##
## Pixbuff work
##

def make_icon_frame(pixbuf, blend=False, border=1, color=0x000000ff):
    """creates a new Pixmap which contains 'pixbuf' with a border at given
    size around it."""
    result = GdkPixbuf.Pixbuf.new(pixbuf.get_colorspace(),
                            True,
                            pixbuf.get_bits_per_sample(),
                            pixbuf.get_width(),
                            pixbuf.get_height())
    result.fill(color)
    if blend:
        pixbuf.composite(result, 0.5, 0.5,
                        pixbuf.get_width(), pixbuf.get_height(),
                        0.5, 0.5,
                        0.5, 0.5,
                        GdkPixbuf.InterpType.NEAREST,
                        256)
    pixbuf.copy_area(border, border,
                    pixbuf.get_width() - (border * 2), pixbuf.get_height() - (border * 2),
                    result,
                    border, border)
    return result

def add_background(pixbuf, color=0xffffffff):
    """ adds a background with given color to the pixbuf and returns the result
    as new Pixbuf"""
    result = GdkPixbuf.Pixbuf.new(pixbuf.get_colorspace(),
                            True,
                            pixbuf.get_bits_per_sample(),
                            pixbuf.get_width(),
                            pixbuf.get_height())
    result.fill(color)
    pixbuf.composite(result, 0, 0,
                        pixbuf.get_width(), pixbuf.get_height(),
                        0, 0,
                        1, 1,
                        GdkPixbuf.InterpType.NEAREST,
                        255)
    return result

def create_opendocument_thumb(path):
    """ extracts the thumbnail of an Opendocument document as pixbuf """
    thumb = tempfile.NamedTemporaryFile()
    try:
        f = zipfile.ZipFile(path, mode="r")
        try:
            thumb.write(f.read("Thumbnails/thumbnail.png"))
        except KeyError:
            thumb.close()
            return None
        finally:
            f.close()
    except IOError:
        thumb.close()
        return None
    thumb.flush()
    thumb.seek(0)
    try:
        pixbuf = GdkPixbuf.Pixbuf.new_from_file(thumb.name)
    except GLib.GError:
        return None # (LP: #650917)
    thumb.close()
    return add_background(pixbuf)

def __crop_pixbuf(pixbuf, x, y, size=SIZE_LARGE):
    """ returns a part of the given pixbuf as new one """
    result = GdkPixbuf.Pixbuf.new(pixbuf.get_colorspace(),
                            True,
                            pixbuf.get_bits_per_sample(),
                            size[0], size[1])
    pixbuf.copy_area(x, y, x+size[0], y+size[1], result, 0, 0)
    return result

def create_text_thumb(gio_file, size=None, threshold=2):
    """ tries to use pygments to get a thumbnail of a text file """
    if pygments is None:
        return None
    try:
        lexer = get_lexer_for_mimetype(gio_file.mime_type)
    except pygments.util.ClassNotFound:
        lexer = get_lexer_for_mimetype("text/plain")
    if chardet:
        lexer.encoding = "chardet"
    thumb = tempfile.NamedTemporaryFile()
    formatter = ImageFormatter(font_name="DejaVu Sans Mono", line_numbers=False, font_size=10)
    # to speed things up only highlight the first 20 lines
    content = "\n".join(gio_file.get_content().split("\n")[:20])
    try:
        content = highlight(content, lexer, formatter)
    except (UnicodeDecodeError, TypeError):
        # we can't create the pixbuf
        return None
    thumb.write(content)
    thumb.flush()
    thumb.seek(0)
    try:
        pixbuf = GdkPixbuf.Pixbuf.new_from_file(thumb.name)
    except GLib.GError:
        return None # (LP: #743125)
    thumb.close()
    if size is not None:
        new_height = None
        new_width = None
        height = pixbuf.get_height()
        width = pixbuf.get_width()
        if width > threshold*size[0]:
            new_width = threshold*size[0]
        if height > threshold*size[1]:
            new_height = threshold*size[1]
        if new_height is not None or new_width is not None:
            pixbuf = __crop_pixbuf(pixbuf, 0, 0, (new_width or width, new_height or height))
    return pixbuf


def new_grayscale_pixbuf(pixbuf):
    """
    Makes a pixbuf grayscale

    :param pixbuf: a :class:`Pixbuf <GdkPixbuf>`

    :returns: a :class:`Pixbuf <GdkPixbuf>`

    """
    pixbuf2 = pixbuf.copy()
    pixbuf.saturate_and_pixelate(pixbuf2, 0.0, False)
    return pixbuf2

def crop_pixbuf(pixbuf, x, y, width, height):
    """
    Crop a pixbuf

    Arguments:
    :param pixbuf: a :class:`Pixbuf <GdkPixbuf>`
    :param x: the x position to crop from in the source
    :param y: the y position to crop from in the source
    :param width: crop width
    :param height: crop height

    :returns: a :class:`Pixbuf <GdkPixbuf>`
    """
    dest_pixbuf = GdkPixbuf.Pixbuf.new(GdkPixbuf.Colorspace.RGB, True, 8, width, height)
    pixbuf.copy_area(x, y, width, height, dest_pixbuf, 0, 0)
    return dest_pixbuf

def scale_to_fill(pixbuf, neww, newh):
    """
    Scales/crops a new pixbuf to a width and height at best fit and returns it

    Arguments:
    :param pixbuf: a :class:`Pixbuf <GdkPixbuf>`
    :param neww: new width of the new pixbuf
    :param newh: a new height of the new pixbuf

    :returns: a :class:`Pixbuf <GdkPixbuf>`
    """
    imagew, imageh = pixbuf.get_width(), pixbuf.get_height()
    if (imagew, imageh) != (neww, newh):
        imageratio = float(imagew) / float(imageh)
        newratio = float(neww) / float(newh)
        if imageratio > newratio:
            transformw = int(round(newh * imageratio))
            pixbuf = pixbuf.scale_simple(transformw, newh, GdkPixbuf.InterpType.BILINEAR)
            pixbuf = crop_pixbuf(pixbuf, 0, 0, neww, newh)
        elif imageratio < newratio:
            transformh = int(round(neww / imageratio))
            pixbuf = pixbuf.scale_simple(neww, transformh, GdkPixbuf.InterpType.BILINEAR)
            pixbuf = crop_pixbuf(pixbuf, 0, 0, neww, newh)
        else:
            pixbuf = pixbuf.scale_simple(neww, newh, GdkPixbuf.InterpType.BILINEAR)
    return pixbuf

def scale_to_center(pixbuf, neww, newh):
    """
    Scales and centers a new pixbuf to a width and height at best fit and returns it

    Arguments:
    :param pixbuf: a :class:`Pixbuf <GdkPixbuf>`
    :param neww: new width of the new pixbuf
    :param newh: new height of the new pixbuf

    :returns: a :class:`Pixbuf <GdkPixbuf>`
    """

    imagew, imageh = pixbuf.get_width(), pixbuf.get_height()
    if (imagew, imageh) == (neww, newh):
        return pixbuf

    dest_pixbuf = GdkPixbuf.Pixbuf.new(GdkPixbuf.Colorspace.RGB, True, 8, neww, newh)
    dest_pixbuf.fill(0xFFFFFFFF)

    # scale down pixbuf if required, preserving aspect ratio
    if imagew > neww or imageh > newh:
        if (imageh * neww) > (imagew * newh):
            imagew = int(0.5 + imagew * (newh / imageh))
            imageh = newh
        else:
            imageh = int(0.5 + imageh * (neww / imagew))
            imagew = neww

        pixbuf = pixbuf.scale_simple(imagew, imageh, GdkPixbuf.InterpType.BILINEAR)

    # center pixbuf in destination pixbuf
    imagew, imageh = pixbuf.get_width(), pixbuf.get_height()
    dest_x = int((neww - imagew)/2)
    dest_y = int((newh - imageh)/2)
    pixbuf.copy_area(0, 0, imagew, imageh, dest_pixbuf, dest_x ,dest_y)

    return dest_pixbuf

class PixbufCache(dict):
    """
    A pixbuf cache dict which stores, loads, and saves pixbufs to a cache and to
    the users filesystem. The naming scheme for thumb files is based on hash()

    There are huge flaws with this object. It does not have a ceiling, and it
    does not remove thumbnails from the file system. Essentially meaning the
    cache directory can grow forever.
    """
    def __init__(self, *args, **kwargs):
        super(PixbufCache, self).__init__()

    def check_cache(self, uri):
        return self[uri]

    def get_buff(self, key):
        thumbpath = os.path.expanduser("~/.cache/gnome-activity-journal/1_" + str(hash(key)))
        if os.path.exists(thumbpath):
            self[key] = (GdkPixbuf.Pixbuf.new_from_file(thumbpath), True)
            return self[key]
        return None

    def __getitem__(self, key):
        if key in self: # FixMe: revisit this
            return super(PixbufCache, self).__getitem__(key)
        return self.get_buff(key)

    def __setitem__(self, key, value):
        (pb, isthumb) = value
        dir_ = os.path.expanduser("~/.cache/gnome-activity-journal/")
        if not os.path.exists(os.path.expanduser("~/.cache/gnome-activity-journal/")):
            os.makedirs(dir_)
        path = dir_ + str(hash(isthumb)) + "_" + str(hash(key))
        if not os.path.exists(path):
            open(path, 'w').close()
            pb.savev(path, "png", [], [])
        return super(PixbufCache, self).__setitem__(key, (pb, isthumb))

    def get_pixbuf_from_uri(self, uri, size=SIZE_LARGE, iconscale=1):
        """
        Returns a pixbuf and True if a thumbnail was found, else False. Uses the
        Pixbuf Cache for thumbnail compatible files. If the pixbuf is a thumb
        it is cached.

        Arguments:
        :param uri: a uri on the disk
        :param size: a size tuple from thumbfactory
        :param iconscale: a factor to reduce icons by (not thumbs)
        :param w: resulting width
        :param h: resulting height

        Warning! This function is in need of a serious clean up.

        :returns: a tuple containing a :class:`Pixbuf <GdkPixbuf>` and bool
        which is True if a thumbnail was found
        """
        try:
            cached = self.check_cache(uri)
        except GLib.GError:
            cached = None
        if cached:
            return cached
        gfile = GioFile.create(uri)
        thumb = True

        if gfile:
            if gfile.has_preview():
                pb = gfile.get_thumbnail(size=size)
                if pb is None and "audio-x-generic" in gfile.icon_names:
                    pb = gfile.get_icon(size=SIZE_NORMAL[0])
                    thumb = False
            else:
                pb = gfile.get_icon(size=SIZE_NORMAL[0])
                thumb = False
        else: pb = None
        if not pb:
            pb = ICON_THEME.lookup_icon(Gtk.STOCK_MISSING_IMAGE, SIZE_NORMAL[0], Gtk.IconLookupFlags.FORCE_SVG).load_icon()
            thumb = False
        if thumb:
            if "audio-x-generic" in gfile.icon_names: thumb = False
            self[uri] = (pb, thumb)
        return pb, thumb

PIXBUFCACHE = PixbufCache()


def get_icon_for_name(name, size):
    """
    return a icon for a name
    """
    size = int(size)
    ICONS[(size, size)]
    if name in ICONS[(size, size)]:
        return ICONS[(size, size)][name]
    info = ICON_THEME.lookup_icon(name, size, Gtk.IconLookupFlags.USE_BUILTIN)
    if not info:
        return None
    location = info.get_filename()
    return get_icon_for_uri(location, size)

def get_icon_for_uri(uri, size):
    if uri in ICONS[(size, size)]:
        return ICONS[(size, size)][uri]
    pixbuf = GdkPixbuf.Pixbuf.new_from_file_at_size(uri, size, size)
    ICONS[(size, size)][uri] = pixbuf
    return pixbuf

def get_icon_from_object_at_uri(uri, size):
    """
    Returns a icon from a event at size

    :param uri: a uri string
    :param size: a int representing the size in pixels of the icon

    :returns: a :class:`Pixbuf <GdkPixbuf>`
    """
    gfile = GioFile.create(uri)
    if gfile:
        pb = gfile.get_icon(size=size)
        if pb:
            return pb
    return False


##
## Other useful methods
##

def get_menu_item_with_stock_id_and_text(stock_id, text):
    item = Gtk.MenuItem.new_with_mnemonic (text)
    return item


def is_command_available(command):
    """
    Checks whether the given command is available, by looking for it in
    the PATH.

    This is useful for ensuring that optional dependencies on external
    applications are fulfilled.
    """
    assert len(" a".split()) == 1, "No arguments are accepted in command"
    for directory in os.environ["PATH"].split(os.pathsep):
        if os.path.exists(os.path.join(directory, command)):
            return True
    return False

def launch_command(command, arguments=None):
    """
    Launches a program as an independent process.
    """
    if not arguments:
        arguments = []
    null = os.open(os.devnull, os.O_RDWR)
    subprocess.Popen([command] + arguments, stdout=null, stderr=null,
                     close_fds=True)

def launch_string_command(command):
    """
    Launches a program as an independent from a string
    """
    command = command.split(" ")
    null = os.open(os.devnull, os.O_RDWR)
    subprocess.Popen(command, stdout=null, stderr=null,
                     close_fds=True)

##
## GioFile
##

GIO_FILES = {}

class GioFile(object):

    def __new__(classtype, *args, **kwargs):
        # Check to see if a __single exists already for this class
        # Compare class types instead of just looking for None so
        # that subclasses will create their own __single objects
        subj = args[0][1][0][0]
        if subj not in GIO_FILES:
            GIO_FILES[subj] = object.__new__(classtype)
        return GIO_FILES[subj]

    @classmethod
    def create(cls, path):
        """ save method to create a GioFile object, if a file does not exist
        None is returned"""
        try:
            if path not in GIO_FILES:
                GIO_FILES[path] = cls(path)
            return GIO_FILES[path]
        except IOError as e:
            return None

    def __init__(self, path):
        self._file_object = Gio.File.new_for_uri(path) #fixme: naming
        self._file_info = self._file_object.query_info(
            "standard::content-type,standard::icon,time::modified",
            Gio.FileQueryInfoFlags.NONE, None)
        self._file_annotation = self._file_object.query_info(
            "metadata::annotation",
            Gio.FileQueryInfoFlags.NONE, None)

    @property
    def mime_type(self):
        return self._file_info.get_attribute_string("standard::content-type")

    @property
    def mtime(self):
        return self._file_info.get_attribute_uint64("time::modified")

    @property
    def basename(self):
        return self._file_object.get_basename()

    @property
    def uri(self):
        return self._file_object.get_uri()

    def get_annotation(self):
        self.refresh_annotation()
        return self._file_annotation.get_attribute_as_string("metadata::annotation")

    def set_annotation(self, annotation):
        self._file_annotation.set_attribute_string("metadata::annotation", annotation)
        self._file_object.set_attributes_from_info(self._file_annotation,
                                                   Gio.FileQueryInfoFlags.NONE,
                                                   None)

    annotation = property(get_annotation, set_annotation)

    def get_content(self):
        f = open(self._file_object.get_path())
        try:
            content = f.read()
        finally:
            f.close()
        return content

    @property
    def icon_names(self):
        try:
            return self._file_info.get_attribute_object("standard::icon").get_names()
        except AttributeError:
            return list()

    def get_thumbnail(self, size=SIZE_NORMAL, border=0):
        assert size in ICON_SIZES
        try:
            thumb = THUMBS[size][self.uri]
        except KeyError:
            factory = THUMBNAIL_FACTORIES[size]
            location = factory.lookup(self.uri, self.mtime)
            if location:
                thumb, mtime = THUMBS[size][self.uri] = \
                    (GdkPixbuf.Pixbuf.new_from_file(location), self.mtime)
            else:
                if factory.has_valid_failed_thumbnail(self.uri, self.mtime):
                    thumb = THUMBS[size][self.uri] = None
                else:
                    thumb = factory.generate_thumbnail(self.uri, self.mime_type)
                    if thumb is None:
                        # maybe we are able to use a custom thumbnailer here
                        if [name for name in self.icon_names if "application-vnd.oasis.opendocument" in name]:
                            thumb = create_opendocument_thumb(self._file_object.get_path())
                        elif "text-x-generic" in self.icon_names or "text-x-script" in self.icon_names:
                            try:
                                thumb = create_text_thumb(self, size, 1)
                            except Exception:
                                thumb = None
                        elif "audio-x-generic" in self.icon_names:
                            pass
                            #FIXME
                            #thumb = self.get_audio_cover(size)
                    if thumb is None:
                        factory.create_failed_thumbnail(self.uri, self.mtime)
                    else:
                        width, height = thumb.get_width(), thumb.get_height()
                        if width > size[0] or height > size[1]:
                            scale = min(float(size[0])/width, float(size[1])/height)
                            thumb = thumb.scale_simple(int(scale*width), int(scale*height), GdkPixbuf.InterpType.BILINEAR)
                        factory.save_thumbnail(thumb, self.uri, self.mtime)
                        THUMBS[size][self.uri] = (thumb, self.mtime)
        else:
            if thumb is not None:
                if thumb[1] != self.mtime:
                    del THUMBS[size][self.uri]
                    return self.get_thumbnail(size, border)
                thumb = thumb[0]
        if thumb is not None and border:
            thumb = make_icon_frame(thumb, border=border, color=0x00000080)
        return thumb

    def get_audio_cover(self, size):
        """
        Try to get a cover art in the folder of the song.
        It's s simple hack, but i think it's a good and quick compromise.
        """
        pix = None
        dirname = os.path.dirname(self.event.subjects[0].uri)
        dirname = urllib.parse.unquote(dirname)[7:]
        if not os.path.exists(dirname): return pix
        for f in os.listdir(dirname):
            if f.endswith(".jpg") or f.endswith(".jpeg"):
                path = dirname + os.sep + f
                pix = GdkPixbuf.Pixbuf.new_from_file(path)
                return pix

        return pix

    @property
    def thumbnail(self):
        return self.get_thumbnail()

    def get_monitor(self):
        return self._file_object.monitor_file()

    def refresh(self):
        self._file_info = self._file_object.query_info(
            "standard::content-type,standard::icon,time::modified",
            Gio.FileQueryInfoFlags.NONE, None)

    def refresh_annotation(self):
        self._file_annotation = self._file_object.query_info(
            "metadata::annotation",
            Gio.FileQueryInfoFlags.NONE, None)

    def get_icon(self, size=24, can_thumb=False, border=0):
        icon = None
        if can_thumb:
            # let's try to find a thumbnail
            # we only allow thumbnails as icons for a few content types
            if self.thumb_icon_allowed():
                if (size,) < SIZE_NORMAL:
                    thumb_size = SIZE_NORMAL
                else:
                    thumb_size = SIZE_LARGE
                thumb = self.get_thumbnail(size=thumb_size)
                if thumb:
                    s = float(size)
                    width = thumb.get_width()
                    height = thumb.get_height()
                    scale = min(s/width, s/height)
                    icon = thumb.scale_simple(int(width*scale), int(height*scale), GdkPixbuf.InterpType.NEAREST)
        if icon is None:
            try:
                return ICONS[size][self.uri]
            except KeyError:
                for name in self.icon_names:
                    info = ICON_THEME.lookup_icon(name, size, Gtk.IconLookupFlags.USE_BUILTIN)
                    if info is None:
                        continue
                    location = info.get_filename()
                    if not location:
                        continue # (LP: #722227)
                    icon = GdkPixbuf.Pixbuf.new_from_file_at_size(location, size, size)
                    if icon:
                        break
        ICONS[size][self.uri] = icon
        if icon is not None and border:
            icon = make_icon_frame(icon, border=border, color=0x00000080)
        return icon

    @property
    def icon(self):
        return self.get_icon()

    def launch(self):
        appinfo = Gio.AppInfo.get_default_for_type(self.mime_type, False)
        appinfo.launch([self._file_object,], None)

    def has_preview(self):
        icon_names = self.icon_names
        is_opendocument = [name for name in icon_names if "application-vnd.oasis.opendocument" in name]
        return "video-x-generic" in icon_names \
            or "image-x-generic" in icon_names \
            or "audio-x-generic" in icon_names \
            or "application-pdf" in icon_names \
            or (("text-x-generic" in icon_names or "text-x-script" in icon_names) and pygments is not None) \
            or is_opendocument

    def thumb_icon_allowed(self):
        icon_names = self.icon_names
        is_opendocument = [name for name in icon_names if "application-vnd.oasis.opendocument" in name]
        return "video-x-generic" in icon_names \
            or "image-x-generic" in icon_names \
            or "application-pdf" in icon_names \
            or is_opendocument

    def __eq__(self, other):
        if not isinstance(other, GioFile):
            return False
        return self.uri == other.uri

class DayParts:

    TIMELABELS = [_("Early Morning"), _("Morning"), _("Afternoon"), _("Evening"), _("Night")]
    CHANGEHOURS = config.CHANGEHOURS

    @classmethod
    def _local_minimum(cls, hour):
        for i in range(1, len(cls.CHANGEHOURS)):
            if hour < cls.CHANGEHOURS[i]:
                return i - 1
        return len(cls.CHANGEHOURS) - 1

    @classmethod
    def get_day_parts(cls):
        return cls.TIMELABELS

    @classmethod
    def get_day_part_for_item(cls, item):
        t = time.localtime(int(item.event.timestamp) / 1000)
        return cls._local_minimum(t.tm_hour)

    @classmethod
    def get_day_part_range_for_item(cls, item):
        start = list(time.localtime(int(item.event.timestamp) / 1000))
        i = cls._local_minimum(start[3])
        start[3] = cls.CHANGEHOURS[i] # Reset hour
        start[4] = start[5] = 0 # Reset minutes and seconds
        end = list(start)
        end[3] = cls.CHANGEHOURS[i+1] if len(cls.CHANGEHOURS) > (i + 1) else 23
        end[4] = end[5] = 59
        return time.mktime(tuple(start)) * 1000, time.mktime(tuple(end)) * 1000

def ignore_exceptions(return_value_on_failure=None):
    def wrap(f):
        def safe_method(*args, **kwargs):
            try:
                return f(*args, **kwargs)
            except Exception as e:
                from traceback import print_exc
                print('\n --- Error running %s ---' % f.__name__)
                print_exc()
                print()
                return return_value_on_failure
        return safe_method
    return wrap

def update_cursor(widget, *args):
    gdk_display = Gdk.Display.get_default()

    if config.IN_ERASE_MODE:
        cursor = Gdk.Cursor.new(Gdk.CursorType.PIRATE)
        if not cursor:
            cursor = Gdk.Cursor.new_from_name(gdk_display, "cell")
    else:
        cursor = Gdk.Cursor.new(Gdk.CursorType.ARROW)
        if not cursor:
            cursor = Gdk.Cursor.new_from_name(gdk_display, "default")

    widget.props.window.set_cursor(cursor)

def get_readable_uri(uri):
    uri = Gio.File.new_for_uri(uri).get_uri()

    if uri.startswith("file:///"):
        uri = uri[7:]

    uri = GLib.uri_unescape_string(uri)

    return uri

def confirm_object_deletion(toplevel, uri, delete_all=False, count=0):
    uri = get_readable_uri(uri)
    uri = uri.replace("&", "&amp;")
    primary = _("Confirm event deletion ?")

    if delete_all:
        secondary = _("Delete all <b>%d</b> events for <b>%s</b> from activity journal") % (count, uri)
    else:
        secondary = _("Delete this event for <b>%s</b> from activity journal") % (uri)

    dialog = Gtk.MessageDialog(toplevel,
                               Gtk.DialogFlags.MODAL,
                               Gtk.MessageType.WARNING,
                               Gtk.ButtonsType.OK_CANCEL,
                               primary)

    if delete_all:
        ok_button = dialog.get_widget_for_response(Gtk.ResponseType.OK)
        sc = ok_button.get_style_context()
        sc.add_class("destructive-action")

    dialog.format_secondary_markup(secondary)
    response = dialog.run()
    dialog.destroy()

    return response == Gtk.ResponseType.OK

def update_default_search_colors():
    label = Gtk.Label()
    style = label.get_style_context()

    # save default text color to gsettings if it's
    # not customized
    if not config.SEARCH_RESULTS_TEXT_COLOR_UPDATED:
        text_rgba = get_gtk_rgba(style, "text", 0)
        color = Gdk.RGBA.to_string(text_rgba)
        config.preference_settings["search-results-text-color"] = color
        config.SEARCH_RESULTS_TEXT_COLOR = color

    if not config.SEARCH_RESULTS_HISTOGRAM_COLOR_UPDATED:
        # save default histogram color to gsettings if
        # it's not customized
        histogram_rgba = get_histogram_default_search_color(style)
        color = Gdk.RGBA.to_string(histogram_rgba)
        config.preference_settings["search-results-histogram-color"] = color
        config.SEARCH_RESULTS_HISTOGRAM_COLOR = color

def style_search_result_text(text):
    if config.SEARCH_RESULTS_USE_BOLD_STYLE:
        text = "<b>" + text + "</b>"
    if config.SEARCH_RESULTS_USE_LARGE_STYLE:
        text = "<span size='large'>" + text + "</span>"
    return text

# simplistic for now
def is_light_theme():
    label = Gtk.Label()
    style = label.get_style_context()
    theme_bg_color = c = get_gtk_rgba(style, "bg", 0)

    f = lambda num: (num * 255 > 200)
    return f(c.red) and f(c.green) and f(c.blue)

